import logging
import numpy as np
from typing import Any
from pathlib import Path
from typing import Optional, List
from pydantic.v1 import BaseModel, Field, validator

from jade.utils.utils import load_data

logger = logging.getLogger(__name__)


class UpgradeParamsBaseModel(BaseModel):
    """Base model for all upgrade cost analysis parameters"""

    class Config:
        title = "UpgradeParamsBaseModel"
        anystr_strip_whitespace = True
        validate_assignment = True
        validate_all = True
        extra = "forbid"
        use_enum_values = False
        arbitrary_types_allowed = True

    @classmethod
    def from_file(cls, filename: Path):
        """Return an instance from a file

        Parameters
        ----------
        filename : Path

        """
        return cls(**load_data(filename))
    

class CommonLineParameters(UpgradeParamsBaseModel):
    """This model contains common line parameters that are used in linecode technical catalog, line technical catalog, line output upgrades. 
    All these fields are directly available from opendss"""
    r1: Any = Field(
        title="r1",
        description="r1",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    x1: Any = Field(
        title="x1",
        description="x1",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    r0: Any = Field(
        title="r0",
        description="r0",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    x0: Any = Field(
        title="x0",
        description="x0",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    C1: Any = Field(
        title="c1",
        description="c1",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    C0: Any = Field(
        title="c0",
        description="c0",
        determine_upgrade_option=True,
        symmetrical_impedance_property=True,
    )
    rmatrix: List[str] = Field(
        title="rmatrix",
        description="rmatrix. If provided, should be a list.",
        determine_upgrade_option=True,
        matrix_impedance_property=True,
    )
    xmatrix: List[str] = Field(
        title="xmatrix",
        description="xmatrix. If provided, should be a list.",
        determine_upgrade_option=True,
        matrix_impedance_property=True,
    )
    cmatrix: List[str] = Field(
        title="cmatrix",
        description="cmatrix. If provided, should be a list.",
        determine_upgrade_option=True,
        matrix_impedance_property=True,
    )
    Rg: float = Field(
        title="Rg",
        description="Rg",
        determine_upgrade_option=True,
        matrix_impedance_property=True,
    )
    Xg: float = Field(
        title="Xg",
        description="Xg",
        determine_upgrade_option=True,
        matrix_impedance_property=True,
    )
    rho: float = Field(
        title="rho",
        description="rho",
        matrix_impedance_property=True,
    )
    B1: Any = Field(
        title="B1",
        description="B1",
        determine_upgrade_option=True,
    )
    B0: Any = Field(
        title="B0",
        description="B0",
        determine_upgrade_option=True,
    )
    normamps: float = Field(
        title="normamps",
        description="normamps",
        determine_upgrade_option=True,
    )
    emergamps: float = Field(
        title="emergamps",
        description="emergamps",
        determine_upgrade_option=True,
    )
    units: str = Field(
        title="units",
        description="units",
        determine_upgrade_option=True,
    )   
    faultrate: Optional[float] = Field(
        title="faultrate",
        description="faultrate",
    )
    pctperm: Optional[float] = Field(
        title="pctperm",
        description="pctperm",
    )
    repair: Optional[float] = Field(
        title="repair",
        description="repair",
    ) 
    Seasons: Optional[Any] = Field(
        title="Seasons",
        description="Seasons",
    )
    Ratings: Optional[Any] = Field(
        title="Ratings",
        description="Ratings",
    ) 
    

class OpenDSSLineParams(CommonLineParameters):

    geometry: Any = Field(
        title="geometry",
        description="geometry",
        determine_upgrade_option=True,
    )
    linecode: Any = Field(
        title="linecode",
        description="If line is defined using linecode, then name of associated line code should be provided here.",
        determine_upgrade_option=True,
    )
    phases: int = Field(
        title="phases",
        description="phases",
        determine_upgrade_option=True,
        deciding_property=True,
    )  
    EarthModel: Any = Field(
        title="EarthModel",
        description="EarthModel",
    )
    cncables: Any = Field(
        title="cncables",
        description="cncables",
    )    
    tscables: Any = Field(
        title="tscables",
        description="tscables",
    )   
    wires: Any = Field(
        title="wires",
        description="wires",
    )
    like: Any = Field(
        title="like",
        description="like",
    )
    basefreq: float = Field(
        title="basefreq",
        description="basefreq",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    Switch: str = Field(
        title="Switch",
        description="Flag that determines whether line is switch or not.",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    spacing: Any = Field(
        title="spacing",
        description="spacing",
        determine_upgrade_option=True,
    )
    
    @validator("Switch")
    def check_switch_property(cls, Switch):
        if Switch not in ("Yes", "No"):
            raise ValueError("Incorrect Switch type.")
        return Switch
    

class ExtraLineParams(BaseModel):
    line_definition_type: str = Field(
        title="line_definition_type",
        description="This indicates if the line is defined by using linecodes or line geometry. Possible values are linecode, geometry or line_definition."
                    "This is a computed field, not a direct OpenDSS object property",
        determine_upgrade_option=True,
    )
    kV: float = Field(
        title="kV",
        description="Line-to-Neutral kV. This is not a direct OpenDSS object property.",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    h: float = Field(
        title="h",
        description="h. This is not a direct opendss line property, and is added as a new field. A value is available if line is defined as a line geometry.",
        determine_upgrade_option=True,
    )
    line_placement: str = Field(
        title="line_placement",
        description="line_placement. This is a new field, not a direct OpenDSS object property.",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    
    @validator("line_definition_type")
    def check_line_definition_type(cls, line_definition_type):
        if line_definition_type not in ("linecode", "geometry", "line_definition"):
            raise ValueError(f"Incorrect Line definition type: {line_definition_type}. Acceptable values: linecode, geometry or line_definition.")
        return line_definition_type
    
    @validator("line_placement")
    def check_line_placement(cls, line_placement):
        if line_placement not in ("underground", "overhead"):
            raise ValueError(f"Incorrect Line placement type: {line_placement}. Acceptable values: overhead, underground.")
        return line_placement


class CommonTransformerParameters(UpgradeParamsBaseModel):
    """This model contains common transformer parameters which are inherited by transformer catalog, and transformer upgrade output."""
    phases: int = Field(
        title="phases",
        description="phases",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    windings: int = Field(
        title="windings",
        description="windings",
        determine_upgrade_option=True,
    )
    wdg: int = Field(
        title="wdg",
        description="wdg",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    conn: str = Field(
        title="conn",
        description="conn",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    kV: float = Field(
        title="kV",
        description="kV",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    kVA: float = Field(
        title="kVA",
        description="kVA",
        determine_upgrade_option=True,
    )
    tap: float = Field(
        title="tap",
        description="tap",
        determine_upgrade_option=True,
    )
    pctR: float = Field(
        title="pctR",
        description="pctR",
        alias="%R",
        determine_upgrade_option=True,
    )
    Rneut: float = Field(
        title="Rneut",
        description="Rneut",
        determine_upgrade_option=True,
    )
    Xneut: float = Field(
        title="Xneut",
        description="Xneut",
        determine_upgrade_option=True,
    )
    conns: List[str] = Field(
        title="conns",
        description="conns. This needs to be passed as a list.",
        determine_upgrade_option=True,
        deciding_property=True,
    )
    kVs: List[float] = Field(
        title="kVs",
        description="kVs. This needs to be passed as a list.",
        determine_upgrade_option=True,
        deciding_property=True,
    )   
    kVAs: List[float] = Field(
        title="kVAs",
        description="kVAs. This needs to be passed as a list.",
        determine_upgrade_option=True,
    )
    taps: List[float] = Field(
        title="taps",
        description="taps. This needs to be passed as a list.",
        determine_upgrade_option=True,
    )
    XHL: float = Field(
        title="XHL",
        description="XHL",
        determine_upgrade_option=True,
    )
    XHT: float = Field(
        title="XHT",
        description="XHT",
        determine_upgrade_option=True,
    )
    XLT: float = Field(
        title="XLT",
        description="XLT",
        determine_upgrade_option=True,
    )
    Xscarray: List[float] = Field(
        title="Xscarray",
        description="Xscarray",
        determine_upgrade_option=True,
    )
    thermal: Any = Field(
        title="thermal",
        description="thermal",
        determine_upgrade_option=True,
        write_property=True,
    )
    n: float = Field(
        title="n",
        description="n",
        determine_upgrade_option=True,
        write_property=True,
    )
    m: float = Field(
        title="m",
        description="m",
        determine_upgrade_option=True,
        write_property=True,
    )
    flrise: float = Field(
        title="flrise",
        description="flrise",
        determine_upgrade_option=True,
        write_property=True,
    )
    hsrise: float = Field(
        title="hsrise",
        description="hsrise",
        determine_upgrade_option=True,
        write_property=True,
    ) 
    pctloadloss: float = Field(
        title="%loadloss",
        description="%loadloss",
        alias="%loadloss",
        determine_upgrade_option=True,
        write_property=True,
    )
    pctnoloadloss: float = Field(
        title="%noloadloss",
        description="%noloadloss",
        alias="%noloadloss",
        determine_upgrade_option=True,
        write_property=True,
    )
    normhkVA: float = Field(
        title="normhkVA",
        description="normhkVA",
        determine_upgrade_option=True,
        write_property=True,
    ) 
    emerghkVA: float = Field(
        title="emerghkVA",
        description="emerghkVA",
        determine_upgrade_option=True,
        write_property=True,
    )
    sub: str = Field(
        title="sub",
        description="sub",
        determine_upgrade_option=True,
    )
    MaxTap: float = Field(
        title="MaxTap",
        description="MaxTap",
        determine_upgrade_option=True,
    ) 
    MinTap: float = Field(
        title="MinTap",
        description="MinTap",
        determine_upgrade_option=True,
    )
    NumTaps: float = Field(
        title="NumTaps",
        description="NumTaps",
        determine_upgrade_option=True,
        write_property=True,
    )
    subname: Any = Field(
        title="subname",
        description="subname",
        determine_upgrade_option=True,
    ) 
    pctimag: float = Field(
        title="%imag",
        description="%imag",
        alias="%imag",
        determine_upgrade_option=True,
        write_property=True,
    )
    ppm_antifloat: float = Field(
        title="ppm_antifloat",
        description="ppm_antifloat",
        determine_upgrade_option=True,
        write_property=True,
    )
    pctRs: List[float] = Field(
        title="%Rs",
        description="%Rs. If present, should be a list.",
        alias="%Rs",
        determine_upgrade_option=True,
    ) 
    bank: Any = Field(
        title="bank",
        description="bank",
        determine_upgrade_option=True,
    )
    XfmrCode: Any = Field(
        title="XfmrCode",
        description="XfmrCode",
        determine_upgrade_option=True,
    )
    XRConst: Any = Field(
        title="XRConst",
        description="XRConst",
        determine_upgrade_option=True,
        write_property=True,
    ) 
    X12: float = Field(
        title="X12",
        description="X12",
        determine_upgrade_option=True,
    )
    X13: float = Field(
        title="X13",
        description="X13",
        determine_upgrade_option=True,
    )
    X23: float = Field(
        title="X23",
        description="X23",
        determine_upgrade_option=True,
    ) 
    LeadLag: Any = Field(
        title="LeadLag",
        description="LeadLag",
        determine_upgrade_option=True,
        deciding_property=True,
        write_property=True,
    )
    Core: Any = Field(
        title="Core",
        description="Core",
        determine_upgrade_option=True,
        write_property=True,
    )
    RdcOhms: float = Field(
        title="RdcOhms",
        description="RdcOhms",
        determine_upgrade_option=True,
    )
    normamps: float = Field(
        title="normamps",
        description="normamps",
        determine_upgrade_option=True,
    )
    emergamps: float = Field(
        title="emergamps",
        description="emergamps",
        determine_upgrade_option=True,
    )
    faultrate: float = Field(
        title="faultrate",
        description="faultrate",
        determine_upgrade_option=True,
        write_property=True,
    )  
    pctperm: float = Field(
        title="pctperm",
        description="pctperm",
        determine_upgrade_option=True,
    )  
    basefreq: float = Field(
        title="basefreq",
        description="basefreq",
        determine_upgrade_option=True,
        deciding_property=True,
    )  
    repair: Optional[float] = Field(
        title="repair",
        description="repair",
    )
    Seasons: Optional[Any] = Field(
        title="Seasons",
        description="Seasons",
    )
    Ratings: Optional[Any] = Field(
        title="Ratings",
        description="Ratings",
    )

    
class ExtraTransformerParams(BaseModel):
    amp_limit_per_phase: float = Field(
        title="amp_limit_per_phase",
        description="amp_limit_per_phase. This is a new field, not a direct OpenDSS object property.",
        determine_upgrade_option=True,
    )



import pathlib

from jade.utils.utils import load_data, dump_data


def set_hc_postprocess_blocked_by(commands_file):
    """Update blocked_by of hosting capacity postprocess job"""
    path = pathlib.Path(commands_file)
    command = path.read_text().split("\n")[0]
    config_file = command.split("=")[-1]
    data = load_data(config_file)
    
    if len(data["jobs"]) <= 1:
        return
    
    for i in range(1, len(data["jobs"])):
        data["jobs"][i]["blocked_by"].append(i)
    dump_data(data, config_file, indent=2)



"""Defines Time Series Impact Analysis object."""
from disco.analysis import Analysis, Input, Option
from disco.utils.custom_type import CustomType


class TimeSeriesImpactAnalysis(Analysis):
    """Snapshot impact analysis class with default values"""
    INPUTS = [
        Input('over_voltage', CustomType(float), 1.05),
        Input('under_voltage', CustomType(float), 0.95),
        Input('over_voltage_conservative', CustomType(float), 1.05833),
        Input('under_voltage_conservative', CustomType(float), 0.91667),
        Input('line_overload_1', CustomType('percent'), 100),
        Input('line_overload_2', CustomType('percent'), 100),
        Input('transformer_overload_1', CustomType('percent'), 100),
        Input('transformer_overload_2', CustomType('percent'), 100),
    ]

    def run(self, output, **kwargs):
        """Run time series impact analysis"""



"""Make new upgrade cost analysis tables for all jobs in a batch."""

import csv
import json
import itertools
import logging
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

import click
import pandas as pd

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.loggers import setup_logging
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.jobs.results_aggregator import ResultsAggregator
from jade.utils.utils import load_data, dump_data

from disco.models.upgrade_cost_analysis_generic_output_model import (
    UpgradeViolationResultModel,
    TotalUpgradeCostsResultModel,
)
from disco.pipelines.utils import ensure_jade_pipeline_output_dir

logger = logging.getLogger(__name__)

JobInfo = namedtuple(
    "JobInfo", ["name", "substation", "feeder", "placement", "sample", "penetration_level"]
)


def parse_batch_results(output_dir):
    """Parse the results from all jobs in a JADE output directory."""
    output_path = Path(output_dir)
    config = create_config_from_file(output_path / CONFIG_FILE)
    jobs = []
    results = ResultsAggregator.list_results(output_dir)
    result_lookup = {x.name: x for x in results}
    for job in config.iter_jobs():
        if job.name not in result_lookup:
            logger.info("Skip missing job %s", job.name)
            continue
        if result_lookup[job.name].return_code != 0:
            logger.info("Skip failed job %s", job.name)
            continue
        jobs.append(job)

    output_json = {
        "results": [],
        "costs_per_equipment": [],
        "violation_summary": [],
        "equipment": [],
    }

    with ProcessPoolExecutor() as executor:
        for tables in executor.map(parse_job_results, jobs, itertools.repeat(output_path)):
            output_json = combine_job_outputs(output_json, tables)

    for key in output_json: 
        if not output_json[key]:
            logger.warning("There were no %s results.", key)
    
    filename = output_path / "upgrade_summary.json"
    dump_data(output_json, filename, indent=2)
    logger.info("Output summary data to %s", filename)
    # serialize_table(upgrade_summary_table, output_path / "upgrade_summary.csv")
    # serialize_table(total_upgrade_costs_table, output_path / "total_upgrade_costs.csv")


def parse_job_results(job, output_path):
    job_path = output_path / JOBS_OUTPUT_DIR / job.name
    deployment = job.model.deployment
    job_info = JobInfo(
        name=job.name,
        substation=deployment.substation,
        feeder=deployment.feeder,
        placement=deployment.project_data["placement"],
        sample=deployment.project_data["sample"],
        penetration_level=deployment.project_data["penetration_level"],
    )
    overall_output_summary_file = job_path / "output.json"
    data = load_data(overall_output_summary_file)
    tables = get_upgrade_tables(data)
    return tables


def get_upgrade_tables(data):
    tables = {}
    # the key "results" is a dict, but all others are lists of dict
    # so convert "results" dict to list
    for key, value in data.items():
        if key == "results":
            tables[key] = [value]
        else:
            tables[key] = value
    return tables


def combine_job_outputs(output_json, tables):    
    # It might seem odd to go from dict to model back to dict, but this validates
    # fields and types.
    for record in tables["violation_summary"]:
        output_json["violation_summary"].append(UpgradeViolationResultModel(**record).dict())
    for record in tables["costs_per_equipment"]:
        output_json["costs_per_equipment"].append(TotalUpgradeCostsResultModel(**record).dict())
    output_json["equipment"] += tables["equipment"]
    output_json["results"] += tables["results"]
    return output_json


def serialize_table(table, filename):
    """Serialize a list of dictionaries to a CSV file."""
    with open(filename, "w") as f:
        assert table, filename
        fields = table[0].keys()
        writer = csv.DictWriter(f, fieldnames=fields)
        writer.writeheader()
        writer.writerows(table)
        logger.info("Wrote %s", filename)


@click.command()
@click.argument("output_dir")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def make_upgrade_tables(output_dir, verbose):
    """Make upgrade cost tables for all jobs in a batch."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])
    output_dir = ensure_jade_pipeline_output_dir(output_dir)
    parse_batch_results(output_dir)



import logging
import os

from jade.jobs.job_configuration_factory import create_config_from_file
from jade.utils.utils import load_data
from PyDSS.controllers import PvControllerModel

from disco.extensions.upgrade_simulation.upgrade_configuration import UpgradeConfiguration
from disco.extensions.upgrade_simulation.upgrade_inputs import UpgradeInputs
from disco.extensions.upgrade_simulation.upgrade_simulation import UpgradeSimulation
from disco.pydss.pydss_configuration_base import DEFAULT_CONTROLLER_CONFIG_FILE
from disco.version import __version__ as disco_version

logger = logging.getLogger(__name__)


def auto_config(inputs, **kwargs):
    """Create a configuration file for automated upgrade simulation/analysis.

    Parameters
    ----------
    inputs : str
        The model-inputs path for automated upgrade simulation.
    
    Returns
    -------
    :obj:`JobConfiguration`
        An instance of job configuration.
    """
    if not os.path.exists(inputs):
        raise FileNotFoundError(f"Inputs path '{inputs}' does not exist.")

    inputs = UpgradeInputs(inputs)
    config = UpgradeConfiguration(inputs=inputs, **kwargs)
    for job in config.inputs.iter_jobs():
        config.add_job(job)

    return config


def run(config_file, name, output, output_format, verbose):
    """Run automated upgrade simulation through command line"""
    os.makedirs(output, exist_ok=True)

    config = create_config_from_file(config_file, do_not_deserialize_jobs=True)
    job = config.get_job(name)

    logger.info("disco version = %s", disco_version)

    simulation = UpgradeSimulation(
        job=job,
        job_global_config=config.job_global_config,
        output=output
    )
    try:
        upgrade_simulation_params = config.job_global_config["upgrade_simulation_params"]
        dc_ac_ratio = upgrade_simulation_params["dc_ac_ratio"]
        enable_pydss_controller = upgrade_simulation_params["enable_pydss_controller"]
        if enable_pydss_controller:
            pv_controllers = load_data(DEFAULT_CONTROLLER_CONFIG_FILE)
            pydss_controller_model = PvControllerModel(
                **pv_controllers[upgrade_simulation_params["pydss_controller_name"]]
            )
        else:
            pv_controllers = None
            pydss_controller_model= None

        thermal_config = config.job_global_config["thermal_upgrade_params"]
        voltage_config = config.job_global_config["voltage_upgrade_params"]
        cost_database_filepath = config.job_global_config["upgrade_cost_database"]
        ret = simulation.run(
            dc_ac_ratio = dc_ac_ratio,
            enable_pydss_solve=enable_pydss_controller,
            pydss_controller_model=pydss_controller_model,
            thermal_config=thermal_config,
            voltage_config=voltage_config,
            cost_database_filepath=cost_database_filepath,
            verbose=verbose
        )
        return ret
    except Exception:
        logger.exception("Unexcepted error in automatic upgrade analysis job=%s", job)
        raise



"""Base functionality for distribution configurations."""

import logging

from jade.jobs.job_configuration import JobConfiguration
from jade.utils.utils import load_data


logger = logging.getLogger(__name__)


class DistributionConfiguration(JobConfiguration):
    """Represents the configuration options for a distribution simulation."""

    @classmethod
    def deserialize(cls, filename_or_data, do_not_deserialize_jobs=False):
        if isinstance(filename_or_data, str):
            data = load_data(filename_or_data)
        else:
            data = filename_or_data
        data["do_not_deserialize_jobs"] = do_not_deserialize_jobs
        return cls(**data)



"""Make hosting capacity summary tables for all jobs in a batch."""

import csv
import itertools
import json
import logging
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

import click

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.jobs.results_aggregator import ResultsAggregator

from PyDSS.node_voltage_metrics import VOLTAGE_METRIC_FIELDS_TO_INCLUDE_AS_PASS_CRITERIA
from PyDSS.pydss_results import PyDssResults, PyDssScenarioResults
from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO
from PyDSS.thermal_metrics import create_summary_from_dict

from disco.pipelines.utils import ensure_jade_pipeline_output_dir
from disco.pydss.common import SCENARIO_NAME_DELIMITER


JobInfo = namedtuple(
    "JobInfo", ["name", "substation", "feeder", "placement", "sample", "penetration_level"]
)
SNAPSHOT_TIME_POINT_MAPPING = {
    "max_pv_load_ratio": "Max PV to Load Ratio",
    "max_load": "Max Load",
    "daytime_min_load": "Min Daytime Load",
    "pv_minus_load": "Max PV minus Load",
}


logger = logging.getLogger(__name__)


def parse_batch_results(output_dir):
    """Parse the results from all jobs in a JADE output directory."""
    output_path = Path(output_dir)
    config = create_config_from_file(output_path / CONFIG_FILE)
    jobs = []
    results = ResultsAggregator.list_results(output_dir)
    result_lookup = {x.name: x for x in results}
    for job in config.iter_pydss_simulation_jobs():
        if job.name not in result_lookup:
            logger.info("Skip missing job %s", job.name)
            continue
        if result_lookup[job.name].return_code != 0:
            logger.info("Skip failed job %s", job.name)
            continue
        jobs.append(job)

    feeder_head_table = []
    feeder_losses_table = []
    metadata_table = []
    thermal_metrics_table = []
    voltage_metrics_table = []
    snapshot_time_points_table = []

    # This will create flattened tables for each metric across jobs and PyDSS scenarios
    # within eah job.
    # Every table contains job name, substation, feeder, etc., and scenario name, as
    # well as the metrics.
    with ProcessPoolExecutor() as executor:
        for result in executor.map(parse_job_results, jobs, itertools.repeat(output_path)):
            feeder_head_table += result[0]
            feeder_losses_table += result[1]
            metadata_table += result[2]
            thermal_metrics_table += result[3]
            voltage_metrics_table += result[4]
            snapshot_time_points_table += result[5]

    serialize_table(feeder_head_table, output_path / "feeder_head_table.csv")
    serialize_table(feeder_losses_table, output_path / "feeder_losses_table.csv")
    serialize_table(metadata_table, output_path / "metadata_table.csv")
    serialize_table(thermal_metrics_table, output_path / "thermal_metrics_table.csv")
    serialize_table(voltage_metrics_table, output_path / "voltage_metrics_table.csv")
    if jobs and jobs[0].model.model_type == "SnapshotImpactAnalysisModel":
        serialize_table(snapshot_time_points_table, output_path / "snapshot_time_points_table.csv")


def parse_job_results(job, output_path):
    """Return the tables for a single job."""
    job_path = output_path / JOBS_OUTPUT_DIR / job.name / "pydss_project"
    deployment = job.model.deployment
    if job.model.is_base_case:
        job_info = JobInfo(
            name=job.name,
            substation=deployment.substation,
            feeder=deployment.feeder,
            placement="",
            sample="",
            penetration_level="",
        )
    else:
        job_info = JobInfo(
            name=job.name,
            substation=deployment.substation,
            feeder=deployment.feeder,
            placement=deployment.project_data.get("placement", "NA"),
            sample=deployment.project_data.get("sample", 0.0),
            penetration_level=deployment.project_data.get("penetration_level", 0.0),
        )
    results = PyDssResults(job_path)
    metadata_table = get_metadata_table(results, job_info)
    feeder_head_table = get_feeder_head_info(results, job_info)
    feeder_losses_table = get_feeder_losses(results, job_info)
    thermal_metrics_table = get_thermal_metrics(results, job_info)
    voltage_metrics_table = get_voltage_metrics(results, job_info)
    if job.model.model_type == "SnapshotImpactAnalysisModel":
        snapshot_time_points_table = get_snapshot_time_points_table(results, job_info)
    else:
        snapshot_time_points_table = []

    return (
        feeder_head_table,
        feeder_losses_table,
        metadata_table,
        thermal_metrics_table,
        voltage_metrics_table,
        snapshot_time_points_table,
    )


def add_scenario(row, scenario_name, include_time_point=True):
    """Splits the PyDSS scenario name into name and time point.
    Only applicable to loadshape-based snapshot simulations.

    """
    if SCENARIO_NAME_DELIMITER not in scenario_name:
        row["scenario"] = scenario_name
        return

    name, time_point = scenario_name.split(SCENARIO_NAME_DELIMITER)
    if name not in (CONTROL_MODE_SCENARIO, PF1_SCENARIO):
        raise Exception(f"Unexpected parsing of scenario name: {scenario_name}")
    row["scenario"] = name
    if include_time_point:
        row["time_point"] = time_point


def get_metadata_table(results: PyDssResults, job_info: JobInfo):
    """Return capacity stats for each scenario in the PyDSS project"""
    metadata_table = []
    for scenario in results.scenarios:
        total_pv = compute_total_pv_kw(scenario)
        total_load = compute_total_load_kw(scenario)
        pct_pv_to_load_ratio = 100 * total_pv / max(total_load, 1e-3)
        row = job_info._asdict()
        row.update(
            {
                "pct_pv_to_load_ratio": pct_pv_to_load_ratio,
                "pv_capacity_kw": total_pv,
                "load_capacity_kw": total_load,
            }
        )
        add_scenario(row, scenario.name, include_time_point=False)
        metadata_table.append(row)

    return metadata_table


def _read_file(results, path):
    try:
        return json.loads(results.read_file(path))
    except KeyError:
        return []


def get_feeder_losses(results: PyDssResults, job_info: JobInfo):
    """Return feeder losses for all scenarios."""
    feeder_losses_table = []
    data = _read_file(results, "Reports/feeder_losses.json")
    if not data:
        return feeder_losses_table
    for scenario, values in data["scenarios"].items():
        row = job_info._asdict()
        add_scenario(row, scenario)
        row.update(values)
        feeder_losses_table.append(row)
    return feeder_losses_table


def get_feeder_head_info(results: PyDssResults, job_info: JobInfo):
    """Return feeder head info for the control_mode scenario."""
    feeder_head_table = []
    for scenario in results.scenarios:
        data = json.loads(results.read_file(f"Exports/{scenario.name}/FeederHeadInfo.json"))
        row = job_info._asdict()
        add_scenario(row, scenario.name)
        row.update(data)
        feeder_head_table.append(row)
    return feeder_head_table


def get_thermal_metrics(results: PyDssResults, job_info: JobInfo):
    """Return the thermal metrics for all scenarios."""
    thermal_metrics_table = []
    data = _read_file(results, "Reports/thermal_metrics.json")
    if not data:
        return thermal_metrics_table
    
    for scenario, metrics in create_summary_from_dict(data).items():
        row = job_info._asdict()
        add_scenario(row, scenario)
        row.update(metrics)
        thermal_metrics_table.append(row)

    return thermal_metrics_table


def get_voltage_metrics(results: PyDssResults, job_info: JobInfo):
    """Return the voltage metrics for all scenarios."""
    voltage_metrics_table = []
    data = _read_file(results, "Reports/voltage_metrics.json")
    if not data:
        return voltage_metrics_table
    
    for scenario in data["scenarios"]:
        for node_type in data["scenarios"][scenario]:
            row = job_info._asdict()
            add_scenario(row, scenario)
            row["node_type"] = node_type
            for field in VOLTAGE_METRIC_FIELDS_TO_INCLUDE_AS_PASS_CRITERIA:
                summary = row[field] = data["scenarios"][scenario][node_type]["summary"]
                if summary is not None:
                    row[field] = data["scenarios"][scenario][node_type]["summary"][field]
            voltage_metrics_table.append(row)

    return voltage_metrics_table


def get_snapshot_time_points_table(results: PyDssResults, job_info: JobInfo):
    """Return the snapshot time points determined by each job."""
    try:
        data = json.loads(results.read_file(f"Exports/snapshot_time_points.json"))
    except KeyError:
        # Time points are only available if load shapes are used.
        return []

    row = {"name": job_info.name}
    for time_point in SNAPSHOT_TIME_POINT_MAPPING:
        timestamp = data.get(SNAPSHOT_TIME_POINT_MAPPING[time_point])
        if timestamp is not None:
            timestamp = timestamp["Timepoints"]
        row[time_point] = timestamp
    return [row]


def compute_total_pv_kw(scenario: PyDssScenarioResults):
    """Return the total PV capacity in kw."""
    df = scenario.read_element_info_file(f"Exports/{scenario.name}/PVSystemsInfo.csv")
    if "Name" not in df.columns or df.Name.isnull().all():
        return 0
    return df["Pmpp"].sum()


def compute_total_load_kw(scenario: PyDssScenarioResults):
    """Return the total load capacity in kw."""
    df = scenario.read_element_info_file(f"Exports/{scenario.name}/LoadsInfo.csv")
    return df["kW"].sum()


def serialize_table(table, filename):
    """Serialize a list of dictionaries to a CSV file."""
    with open(filename, "w") as f:
        if not table:
            logger.info("No data to write to %s", filename)
            return
        fields = table[0].keys()
        writer = csv.DictWriter(f, fieldnames=fields)
        writer.writeheader()
        writer.writerows(table)
        logger.info("Wrote %s", filename)


@click.command()
@click.argument("output_dir")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def make_summary_tables(output_dir, verbose):
    """Make hosting capacity summary tables for all jobs in a batch."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])
    output_dir = ensure_jade_pipeline_output_dir(output_dir)
    parse_batch_results(output_dir)



"""Common functions for CLI scripts"""

import logging
import logging.config
import os
import shutil

import click

from jade.exceptions import InvalidParameter, UserAbort
from disco.enums import Scale, Placement


logger = logging.getLogger(__name__)


_ENUM_MAPPING = {
    # class to type of value
    "placement": (Placement, str),
    "scale": (Scale, str)
}


def handle_enum_input(_, param, value):
    """Converts inputs to enums."""
    try:
        tup = _ENUM_MAPPING.get(param.name)
        assert tup is not None, "Must add {} to _ENUM_MAPPING".format(
            param.name)
        # Force the correct type onto the value.
        return tup[0](tup[1](value))
    except ValueError as err:
        raise click.BadParameter(str(err))


def _handle_simulation_scripts(ctx, _, value):
    if not os.path.exists(value):
        logger.info("Creating %s...", value)
        try:
            os.makedirs(value)
        except Exception as err:
            logger.error("Could not create %s: %s", value, err)
            raise
    else:
        proceed_with_user_permission(ctx, "Overwrite {}".format(value))
        try:
            scripts = [f for f in os.listdir(value) if f.endswith(".sh")]
            for s in scripts:
                os.remove(os.path.join(value, s))
        except Exception as err:
            logger.error("Could not delete %s: %s", value, err)
            raise

    return value


def handle_existing_dir(directory, force):
    """Deletes existing directory if force is True."""
    if os.path.exists(directory):
        if force:
            shutil.rmtree(directory)
            os.mkdir(directory)
        else:
            raise InvalidParameter(f"directory={directory} exists. Set --force to overwrite")


def proceed_with_user_permission(ctx, message):
    """Pauses execution to prompt user for permission to proceed."""
    prompt = ctx.params.get("prompt")
    if prompt is None:
        prompt = ctx.parent.params["prompt"]
    if not prompt:
        logger.info("Proceed because user passed --no-prompt.")
        return

    answer = ""
    while answer not in ("y", "n"):
        answer = input("{} [Y/N]? ".format(message)).lower()

    if answer == "n":
        raise UserAbort



"""Defines custom types."""

from jade.utils.utils import interpret_datetime


class CustomType:
    """Custom type class allows for additional types outside of str, float, int, etc"""
    CUSTOM_TYPES = [
        'percent',
        'float',
        'int',
        'str',
        'list',
        'dict',
        'bool',
        'datetime',
    ]

    def __init__(self, expected_type):
        self._check_type(expected_type)
        self.expected_type = expected_type

    def get_type_name(self):
        """Either returns string type name or class name of CustomType"""
        if isinstance(self.expected_type, str):
            return self.expected_type

        return self.expected_type.__name__

    def parse_value(self, value):
        """Checks type against given type class and returns cast value

        Parameters
        ----------
        value : str

        Returns
        -------
        casted type or None

        """
        if isinstance(self.expected_type, str):
            return self._handle_custom(value)

        if self.expected_type is bool:
            return self._handle_bool(value)

        try:
            return self.expected_type(value)
        except ValueError:
            return None

    def _check_type(self, given_type):
        """Check if type is a valid custom type

        Parameters
        ----------
        given_type : cls | str

        Raises
        ------
        ValueError
            Raised if type is not a valid custom type

        """
        error = None

        if isinstance(given_type, str):
            if f"{given_type}" not in self.CUSTOM_TYPES:
                error = f"Invalid type: {given_type}"
        else:
            if f"{given_type.__name__}" not in self.CUSTOM_TYPES:
                error = f"Invalid type: {given_type}"

        if error is not None:
            raise ValueError(error)

    def _handle_custom(self, value):
        """Handle all custom types

        Parameters
        ----------
        value : str

        Returns
        -------
        casted type or None

        """
        if self.expected_type == 'percent':
            return float(value)
        if self.expected_type == 'datetime':
            return interpret_datetime(value)

        return None

    @staticmethod
    def _handle_bool(value):
        """Handle bool-like values,
        For example: True/False, TRUE/FALSE, true/false, 1/0

        Parameters
        ----------
        value : str

        Returns
        -------
        casted type or None

        """
        if isinstance(value, bool):
            return value

        if value.lower() in ["true", "1", 1]:
            return True

        if value.lower() in ["false", "0", 0]:
            return False

        return None



"""Common definitions for PyDSS-based simulations."""

import enum
import logging


from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO

logger = logging.getLogger(__name__)


SCENARIO_NAME_DELIMITER = "__"
SIMULATION_POSTPROCESS = "post_process"
TIME_SERIES_SCENARIOS = [CONTROL_MODE_SCENARIO, PF1_SCENARIO]
SNAPSHOT_SCENARIO = "scenario"


class ConfigType(enum.Enum):
    """Possible values for PyDSS config options"""
    SIMULATION_CONFIG = "Simulation"
    CONTROLLER_CONFIG = "PvControllers"
    SCENARIOS = "Scenarios"



"""Model for GEM feeder input data.
The model can be transformed into DISCO OpenDSS model with PV deployments.
"""

import logging
import os
import shutil

import click

from jade.exceptions import InvalidParameter
from disco.cli.common import handle_existing_dir
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.sources.base import BaseOpenDssModel, DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS
from .factory import read_config_data


logger = logging.getLogger(__name__)


COMMON_OPTIONS = (
    click.option(
        "-F",
        "--force",
        help="overwrite existing directory",
        is_flag=True,
        default=False,
        show_default=True,
    ),
)

def common_options(func):
    for option in reversed(COMMON_OPTIONS):
        func = option(func)
    return func


@click.command()
@common_options
@click.option(
    "-o",
    "--output",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["output_dir"],
    show_default=True,
    help="output directory",
)
@click.pass_context
def snapshot(ctx, force, output):
    """Transform input data for a snapshot impact analysis simulation"""
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    GemModel.transform(
        input_file=input_path,
        output_path=output,
        simulation_model=SnapshotImpactAnalysisModel,
    )
    print(f"Transformed data from {input_path} to {output} for Snapshot Analysis.")


class GemModel(BaseOpenDssModel):
    """GEM Feeder Model Inputs Class"""

    TRANSFORM_SUBCOMMANDS = {
        "snapshot": snapshot
    }

    @property
    def substation(self):
        return None

    @staticmethod
    def get_transform_subcommand(name):
        """Return a click command for name."""
        if name not in GemModel.TRANSFORM_SUBCOMMANDS:
            raise InvalidParameter(f"{name} is not supported")
        return GemModel.TRANSFORM_SUBCOMMANDS[name]

    @staticmethod
    def list_transform_subcommands():
        return sorted(list(GemModel.TRANSFORM_SUBCOMMANDS.keys()))

    @classmethod
    def transform(
        cls, input_file, output_path, simulation_model, include_pv_systems=True
    ):
        """Transform GEM input data to a DISCO data model.

        Parameters
        ----------
        input_file : str
        output_path : str

        """
        if os.path.exists(output_path):
            shutil.rmtree(output_path)
        os.makedirs(output_path)
        read_config_data(input_file).generate_output_data(
            output_path, include_pv_systems, simulation_model
        )



import os

from jade.common import OUTPUT_DIR
from disco import timer_stats_collector

from .upgrades.automated_thermal_upgrades import determine_thermal_upgrades
from .upgrades.automated_voltage_upgrades import determine_voltage_upgrades
from .upgrades.cost_computation import compute_all_costs


class UpgradeSimulation:
    
    def __init__(self, job, job_global_config, output=OUTPUT_DIR):
        self.job = job
        self.job_global_config = job_global_config
        self.output = output
    
    @property
    def model(self):
        return self.job.model
    
    @property
    def job_output(self):
        return os.path.join(self.output, self.model.name)
    
    def get_thermal_upgrades_directory(self):
        thermal_upgrades = os.path.join(self.job_output, "ThermalUpgrades")
        os.makedirs(thermal_upgrades, exist_ok=True)
        return thermal_upgrades
    
    def get_voltage_upgrades_directory(self):
        voltage_upgrades = os.path.join(self.job_output, "VoltageUpgrades")
        os.makedirs(voltage_upgrades, exist_ok=True)
        return voltage_upgrades

    def get_upgrade_costs_directory(self):
        upgrade_costs = os.path.join(self.job_output, "UpgradeCosts")
        os.makedirs(upgrade_costs, exist_ok=True)
        return upgrade_costs
    
    def get_feeder_stats_json_file(self):
        feeder_stats_json_file = os.path.join(self.job_output, "feeder_stats.json")
        return feeder_stats_json_file
    
    def get_thermal_upgrades_dss_file(self):
        return os.path.join(self.job_output, "thermal_upgrades.dss")
    
    def get_voltage_upgrades_dss_file(self):
        return os.path.join(self.job_output, "voltage_upgrades.dss")
    
    def get_upgraded_master_dss_file(self):
        return os.path.join(self.job_output, "upgraded_master.dss")

    def get_thermal_upgrades_json_file(self):
        thermal_upgrades = self.get_thermal_upgrades_directory()
        return os.path.join(thermal_upgrades, "thermal_upgrades.json")
    
    def get_voltage_upgrades_json_file(self):
        voltage_upgrades = self.get_voltage_upgrades_directory()
        return os.path.join(voltage_upgrades, "voltage_upgrades.json")
    
    def get_thermal_summary_json_file(self):
        thermal_upgrades = self.get_thermal_upgrades_directory()
        return os.path.join(thermal_upgrades, "thermal_violation_summary.json")
    
    def get_voltage_summary_json_file(self):
        voltage_upgrades = self.get_voltage_upgrades_directory()
        return os.path.join(voltage_upgrades, "voltage_violation_summary.json")

    def internal_upgrades_technical_catalog_filepath(self):
        thermal_upgrades = self.get_thermal_upgrades_directory()
        return os.path.join(thermal_upgrades, "internal_upgrades_technical_catalog.json")
    
    def get_equipment_upgrade_costs_file(self):
        upgrade_costs = self.get_upgrade_costs_directory()
        return os.path.join(upgrade_costs, "equipment_upgrade_costs.json")
    
    def get_voltage_upgrade_costs_file(self):
        upgrade_costs = self.get_upgrade_costs_directory()
        return os.path.join(upgrade_costs, "voltage_upgrade_costs.json")
    
    def get_total_upgrade_costs_file(self):
        upgrade_costs = self.get_upgrade_costs_directory()
        return os.path.join(upgrade_costs, "total_upgrade_costs.json")
    
    def get_overall_output_summary_file(self):
        return os.path.join(self.job_output, "output.json")
    
    @staticmethod
    def generate_command(job, output, config_file, verbose=False):
        """
        Parameters
        ----------
        job: UpgradeParameters
            The instance of upgradeParameters
        output: str
            The output directory of jobs, e.g output/job-outputs
        config_file: str
            The path of config file during runtime
        verbose: bool
            Enable verbose logging if True
        """
        command = [
            "jade-internal run upgrade_simulation",
            f"--name={job.name}",
            f"--output={output}",
            f"--config-file={config_file}"
        ]

        if verbose:
            command.append("--verbose")
        
        return " ".join(command)

    def run(
        self,
        enable_pydss_solve,
        pydss_controller_model,
        dc_ac_ratio,
        thermal_config,
        voltage_config,
        cost_database_filepath,
        verbose=False
    ):  
        determine_thermal_upgrades(
            job_name = self.job.name,
            master_path=self.model.deployment.deployment_file,
            enable_pydss_solve=enable_pydss_solve,
            thermal_config=thermal_config,
            pydss_volt_var_model=pydss_controller_model,
            internal_upgrades_technical_catalog_filepath=self.internal_upgrades_technical_catalog_filepath(),
            thermal_upgrades_dss_filepath=self.get_thermal_upgrades_dss_file(),
            upgraded_master_dss_filepath=self.get_upgraded_master_dss_file(),
            output_json_thermal_upgrades_filepath=self.get_thermal_upgrades_json_file(),
            feeder_stats_json_file = self.get_feeder_stats_json_file(),
            thermal_upgrades_directory=self.get_thermal_upgrades_directory(),
            overall_output_summary_filepath=self.get_overall_output_summary_file(),
            dc_ac_ratio=dc_ac_ratio,
            verbose=verbose
        )
        determine_voltage_upgrades(
            job_name = self.job.name,
            master_path=self.model.deployment.deployment_file,
            enable_pydss_solve=enable_pydss_solve,
            pydss_volt_var_model=pydss_controller_model,
            thermal_config=thermal_config,
            voltage_config=voltage_config,
            thermal_upgrades_dss_filepath=self.get_thermal_upgrades_dss_file(),
            voltage_upgrades_dss_filepath=self.get_voltage_upgrades_dss_file(),
            upgraded_master_dss_filepath=self.get_upgraded_master_dss_file(),
            output_json_voltage_upgrades_filepath = self.get_voltage_upgrades_json_file(),
            feeder_stats_json_file = self.get_feeder_stats_json_file(),
            voltage_upgrades_directory=self.get_voltage_upgrades_directory(),
            overall_output_summary_filepath=self.get_overall_output_summary_file(),
            dc_ac_ratio=dc_ac_ratio,
            verbose=verbose
        )
        compute_all_costs(
            job_name = self.job.name,
            output_json_thermal_upgrades_filepath=self.get_thermal_upgrades_json_file(),
            output_json_voltage_upgrades_filepath=self.get_voltage_upgrades_json_file(),
            cost_database_filepath=cost_database_filepath,
            output_equipment_upgrade_costs_filepath=self.get_equipment_upgrade_costs_file(),
            output_total_upgrade_costs_filepath=self.get_total_upgrade_costs_file(),
            overall_output_summary_filepath=self.get_overall_output_summary_file(),
            feeder_stats_json_file = self.get_feeder_stats_json_file()
        )
        timer_stats_collector.log_stats(clear=True)



"""Common definitions and functionality for tests"""


CONFIG_FILE = "test-config.json"
PRESCREEN_CONFIG_FILE = "test-prescreen-config.json"
PRESCREEN_FINAL_CONFIG_FILE = "test-prescreen-final-config.json"
CONFIG_JOBS = "disco config"
JOB_OUTPUTS = "job-outputs"
MODELS_DIR = "test-disco-models"
OUTPUT = "test-output"
PIPELINE_CONFIG = "test-pipeline.toml"
POST_PROCESS_OUTPUT_FILE = "post-process.toml"
PRESCREEN_JOBS = "disco prescreen-pv-penetration-levels"
SUBMIT_JOBS = "jade submit-jobs --local"
TRANSFORM_MODEL = "disco transform-model"
TRANSFORM_MODEL_LOG = "transform_model.log"
POSTPROCESS_RESULTS = [
    "feeder_head_table.csv",
    "feeder_losses_table.csv",
    "metadata_table.csv",
    "thermal_metrics_table.csv",
    "voltage_metrics_table.csv"
]
UPGRADE_SUMMARY = "upgrade_summary.json"


def find_non_base_case_job(jobs):
    """Return a job that is not a base case."""
    for job in jobs:
        if job.model.base_case is not None:
            assert not job.model.is_base_case
            return job
    raise Exception("Did not find a non-base-case job")



import enum
import logging
from typing import List, Optional, Set, Dict, Any
from pydantic.v1 import Field

from disco.models.upgrade_cost_analysis_equipment_model import UpgradeParamsBaseModel, ExtraLineParams, ExtraTransformerParams
from disco.models.upgrade_cost_analysis_generic_input_model import OpenDSSLineModel, OpenDSSTransformerModel

logger = logging.getLogger(__name__)


class UpgradeViolationResultModel(UpgradeParamsBaseModel):
    """Defines result parameters for thermal upgrades."""

    name: str = Field(
        title="name",
        description="Job name",
    )
    scenario: str = Field(
        title="scenario",
        description="Simulation scenario describing the controls being used",
        default="control_mode",
    )
    stage: str = Field(
        title="stage",
        description="Stage of upgrades: initial (before upgrades) or final (after upgrades)",
    )
    upgrade_type: str = Field(
        title="upgrade_type",
        description="Type of upgrade: thermal or voltage",
    )
    simulation_time_s: float = Field(
        title="simulation_time_s",
        description="Simulation time to perform upgrades (seconds)",
    )
    thermal_violations_present: bool = Field(
        title="thermal_violations_present",
        description="Flag indicating whether thermal violations are present",
    )
    voltage_violations_present: bool = Field(
        title="voltage_violations_present",
        description="Flag indicating whether voltage violations are present",
    )
    max_bus_voltage: float = Field(
        title="max_bus_voltage",
        description="Maximum voltage recorded on any bus",
        units="pu",
    )
    min_bus_voltage: float = Field(
        title="min_bus_voltage",
        description="Minimum voltage recorded on any bus",
        units="pu",
    )
    num_voltage_violation_buses: int = Field(
        title="num_voltage_violation_buses",
        description="Number of buses with voltage violations",
    )
    num_overvoltage_violation_buses: int = Field(
        title="num_overvoltage_violation_buses",
        description="Number of buses with voltage above voltage_upper_limit",
    )
    voltage_upper_limit: float = Field(
        title="voltage_upper_limit",
        description="Voltage upper limit, the threshold considered for determining overvoltages",
        units="pu",
    )
    num_undervoltage_violation_buses: int = Field(
        title="num_undervoltage_violation_buses",
        description="Number of buses with voltage below voltage_lower_limit",
    )
    voltage_lower_limit: float = Field(
        title="voltage_lower_limit",
        description="Voltage lower limit, the threshold considered for determining undervoltages",
        units="pu",
    )
    max_line_loading: float = Field(
        title="max_line_loading",
        description="Maximum line loading",
        units="pu",
    )
    max_transformer_loading: float = Field(
        title="max_transformer_loading",
        description="Maximum transformer loading",
        units="pu",
    )
    num_line_violations: int = Field(
        title="num_line_violations",
        description="Number of lines with loading above line upper limit",
    )
    line_upper_limit: float = Field(
        title="line_upper_limit",
        description="Line upper limit, the threshold considered for determining line overloading",
        units="pu",
    )
    num_transformer_violations: int = Field(
        title="num_transformer_violations",
        description="Number of transformers with loading above transformer upper limit",
    )
    transformer_upper_limit: float = Field(
        title="transformer_upper_limit",
        description="Transformer upper limit, the threshold considered for determining transformer overloading",
        units="pu",
    )


class TotalUpgradeCostsResultModel(UpgradeParamsBaseModel):
    """Provides total output costs for upgrading a type of equipment."""

    name: str = Field(
        title="name",
        description="Job name",
    )
    type: str = Field(
        title="type",
        description="Equipment type",
    )
    count: int = Field(
        title="count",
        description="Count of upgraded equipment",
    )
    total_cost_usd: float = Field(
        title="total_cost_usd",
        description="Total cost in US dollars",
        units="dollars",
    )
    

class EquipmentTypeUpgradeCostsResultModel(UpgradeParamsBaseModel):
    """Provides output costs and final equipment details for each upgraded asset."""

    name: str = Field(
        title="name",
        description="Job name",
    )
    type: str = Field(
        title="type",
        description="Equipment type",
    )
    count: int = Field(
        title="count",
        description="Count of upgraded equipment",
    )
    total_cost_usd: float = Field(
        title="total_cost_usd",
        description="Total cost in US dollars",
        units="dollars",
    )
    equipment_parameters: Optional[Any] = Field(
        title="equipment_parameters",
        description="Final Equipment parameters",
        units="",
        default={},
    )
    comment: Optional[Any] = Field(
        title="comment",
        description="Other comments",
        units="",
        default="",
    )
    

class AllEquipmentUpgradeCostsResultModel(UpgradeParamsBaseModel):
    """Contains outputs for thermal and voltage costs by individual equipment"""
    voltage: Optional[List[EquipmentTypeUpgradeCostsResultModel]] = Field(
        title="voltage", 
        description="Voltage upgrade costs",
        default=[],  
    )
    thermal: Optional[List[EquipmentTypeUpgradeCostsResultModel]] = Field(
        title="thermal", 
        description="Line and Transformer upgrade costs",  
        default=[],
    )

    
class VoltageUpgradesTechnicalResultModel(UpgradeParamsBaseModel):
    """Model for voltage_upgrades.json. This contains the details on the voltages upgrades made to resolve voltage violations"""
    equipment_type: str = Field(
        title="equipment_type",
        description="Type of upgraded equipment",
    )  
    name: str = Field(
        title="name",
        description="Name of the upgraded equipment",
    )  
    new_controller_added: bool = Field(
        title="new_controller_added",
        description="This flag depicts whether a new controller was added",
    ) 
    controller_settings_modified: bool = Field(
        title="controller_settings_modified",
        description="This flag depicts whether the controller settings were modified",
    ) 
    final_settings: dict = Field(
        title="final_settings",
        description="Final Settings of the equipment",
    ) 
    new_transformer_added: bool = Field(
        title="new_transformer_added",
        description="This flag depicts whether a new transformer was added",
    ) 
    at_substation: bool = Field(
        title="at_substation",
        description="This flag depicts whether the change was made at the Substation",
    ) 
       

class LineUpgradesTechnicalResultModel(OpenDSSLineModel, ExtraLineParams):
    """Line Upgrades Output Details model"""
    equipment_type: str = Field(
         title="equipment_type",
        description="Equipment Type"
    )
    upgrade_type: str = Field(
         title="upgrade_type",
        description="Upgrade Type"
    )
    parameter_type: str = Field(
         title="parameter_type",
        description="Parameter Type"
    )
    action: str = Field(
         title="action",
        description="Action"
    )
    final_equipment_name: str = Field(
         title="final_equipment_name",
        description="final_equipment_name"
    )
    original_equipment_name: str = Field(
         title="original_equipment_name",
        description="original_equipment_name"
    )

    
class TransformerUpgradesTechnicalResultModel(OpenDSSTransformerModel, ExtraTransformerParams):
    """Transformer Upgrades Output Details model"""
    equipment_type: str = Field(
         title="equipment_type",
        description="Equipment Type"
    )
    upgrade_type: str = Field(
         title="upgrade_type",
        description="Upgrade Type"
    )
    parameter_type: str = Field(
         title="parameter_type",
        description="Parameter Type"
    )
    action: str = Field(
         title="action",
        description="Action"
    )
    final_equipment_name: str = Field(
         title="final_equipment_name",
        description="final_equipment_name"
    )
    original_equipment_name: str = Field(
         title="original_equipment_name",
        description="original_equipment_name"
    )

    
class AllUpgradesTechnicalResultModel(UpgradeParamsBaseModel):
    """Contains All Upgrades Output Details. Read in as input for cost computation"""
    line: Optional[List[LineUpgradesTechnicalResultModel]] = Field(
        title="line",
        description="line upgrades output details",
        default=[]
    )
    transformer: Optional[List[TransformerUpgradesTechnicalResultModel]] = Field(
        title="transformer",
        description="transformer upgrades output details",
        default=[]
    )
    voltage: Optional[List[VoltageUpgradesTechnicalResultModel]] = Field(
        title="voltage",
        description="voltage upgrades output details",
        default=[]
    )


class UpgradesCostResultSummaryModel(UpgradeParamsBaseModel):
    """Contains individual equipment output"""
    name: str = Field(
        title="name",
        description="Job name",
    )
    equipment_type: str = Field(
        title="equipment_type",
        description="Type of equipment",
    )
    equipment_name: str = Field(
        title="equipment_name",
        description="Name of equipment",
    )
    status: str = Field(
        title="status",
        description="Status",
    )
    total_cost_usd: float = Field(
        title="total_cost_usd",
        description="Total cost in US dollars",
    )
    parameter1_name: str = Field(
        title="parameter1_name",
        description="Name of parameter1",
    )
    parameter1_original: Any = Field(
        title="parameter1_original",
        description="Original value of parameter1",
    )
    parameter1_upgraded: Any = Field(
        title="parameter1_upgraded",
        description="Upgraded value of parameter1",
    )
    parameter2_name: Optional[str] = Field(
        title="parameter2_name",
        description="Name of parameter2",
        default="",
    )
    parameter2_original: Optional[Any] = Field(
        title="parameter2_original",
        description="Original value of parameter2",
        default=None,
    )
    parameter2_upgraded: Optional[Any] = Field(
        title="parameter2_upgraded",
        description="Upgraded value of parameter2",
        default=None,
    )
    parameter3_name: Optional[str] = Field(
        title="parameter3_name",
        description="Name of parameter3",
        default="",
    )
    parameter3_original: Optional[Any] = Field(
        title="parameter3_original",
        description="Original value of parameter3",
        default=None,
    )
    parameter3_upgraded: Optional[Any] = Field(
        title="parameter3_upgraded",
        description="Upgraded value of parameter3",
        default=None,
    )    


class ResultSummaryModel(UpgradeParamsBaseModel):
    """Contains final results summary"""
    total_cost: str = Field(
        title="equipment_type",
        description="Type of equipment",
    )
    equipment_name: str = Field(
        title="equipment_name",
        description="Name of equipment",
    )

class AllUpgradesCostResultSummaryModel(UpgradeParamsBaseModel):
    """Contains All Equipment output"""
    equipment: List[UpgradesCostResultSummaryModel] = Field(
        title="equipment",
        description="Individual Equipment output details",
        default=[]
    )
    

class ExtendedEnum(enum.Enum):
    @classmethod
    def list_values(cls):
        return list(map(lambda c: c.value, cls))
    

class CapacitorControllerResultType(ExtendedEnum):
    """Possible values for capacitor upgrade type"""
    add_new_cap_controller = "Capacitor controller"
    change_cap_control = "Capacitor controller setting change"
   
 
class VoltageRegulatorResultType(ExtendedEnum):
    """Possible values for voltage regulator upgrade type"""
    add_new_reg_control = "In-line voltage regulator"
    change_reg_control = "In-line voltage regulator control setting change"
    # replace_reg_control = "Replace in-line voltage regulator controller"   # this is not used currently
    add_substation_ltc = "Substation LTC"
    change_ltc_control = "Substation LTC setting change"
    add_new_vreg_transformer = "Transformer for voltage regulator"
    add_new_substation_transformer =  "Substation transformer"


class EquipmentUpgradeStatusModel(ExtendedEnum):
    """Possible values for upgrade status in output summary file"""
    replaced = "replaced"
    new = "new"
    setting_changed = "setting_changed"
    unchanged = "unchanged"


class UpgradeJobOutputModel(UpgradeParamsBaseModel):
    """Contains outputs from one job."""

    upgraded_opendss_model_file: str = Field(
        title="upgraded_opendss_model_file",
        description="Path to file that will load the upgraded network.",
    )
    feeder_stats: str = Field(
        title="feeder_stats",
        description="Path to file containing feeder metadata and equipment details before and "
        "after upgrades.",
    )
    return_code: int = Field(
        title="return_code",
        description="Return code from process. Zero is success, non-zero is a failure.",
    )


class UpgradeSimulationOutputModel(UpgradeParamsBaseModel):
    """Contains outputs from all jobs in the simulation."""

    log_file: str = Field(
        title="log_file",
        description="Path to log file for the simulation.",
    )
    jobs: List[UpgradeJobOutputModel] = Field(
        title="jobs",
        description="Outputs for each job in the simulation.",
    )


class JobUpgradeSummaryOutputModel(UpgradeParamsBaseModel):
    """Contains results from all jobs in the simulation."""
    results: List = Field(
        title="results",
        description="Results summary for each job",
    )
    outputs: UpgradeSimulationOutputModel = Field(
        title="outputs",
        description="Outputs for each job in the simulation.",
    )
    violation_summary: List[UpgradeViolationResultModel] = Field(
        title="upgrade_summary",
        description="Contains thermal or voltage upgrade results for each job",
    )
    costs_per_equipment: List[TotalUpgradeCostsResultModel] = Field(
        title="costs_per_equipment",
        description="Contains upgrade cost information for each job by equipment type",
    )    
    equipment: List = Field(
        title="equipment",
        description="Contains equipment information for each job",
    )



"""Contains functionality to configure PyDss simulations."""

import copy
import logging
import os

from jade.exceptions import InvalidParameter
from PyDSS.common import ControllerType
from PyDSS.registry import Registry

import disco
from disco.distribution.distribution_configuration import DistributionConfiguration
from disco.enums import AnalysisType, get_enum_from_value, SimulationType
from disco.pydss.common import ConfigType


logger = logging.getLogger(__name__)

DEFAULT_CONTROLLER_CONFIG_FILE = os.path.join(
    os.path.dirname(getattr(disco, "__path__")[0]), "disco", "pydss",
    "config", "pv_controllers.toml"
)

DEFAULT_CONTROLLER_CONFIGS = [
    {
        "controller_type": ControllerType.PV_CONTROLLER.value,
        "name": "volt_var_1",
        "filename": DEFAULT_CONTROLLER_CONFIG_FILE
    },
    {
        "controller_type": ControllerType.PV_CONTROLLER.value,
        "name": "volt_var_ieee_1547_2018_catB",
        "filename": DEFAULT_CONTROLLER_CONFIG_FILE
    },
]

DEFAULT_LOAD_SHAPE_START_TIME = "2021-01-01 00:00:00.0"

DEFAULT_PYDSS_SIMULATION_CONFIG = {
    "project": {
        "start_time": "2021-01-01 00:00:00.0",
        "loadshape_start_time": DEFAULT_LOAD_SHAPE_START_TIME,
        "simulation_duration_min": 1.0,
        "step_resolution_sec": 900.0,
        "max_control_iterations": 50,
        "error_tolerance": 0.0001,
        "max_error_tolerance": 0.01,
        "convergence_error_percent_threshold": 1.0,
        "skip_export_on_convergence_error": True,
        "use_controller_registry": True
    },
    "exports": {
        "export_results": True,
        "export_elements": True,
        "export_element_types": ["Loads", "PVSystems"],  # Make this empty to export all types.
        "export_event_log": True,
        "export_format": "h5",
        "export_compression": True,
        "export_data_tables": False,
        "export_pv_profiles": False,
    },
    "reports": {},
}

DEFAULT_PYDSS_CONFIG = {
    ConfigType.SIMULATION_CONFIG: {
        "default": copy.deepcopy(DEFAULT_PYDSS_SIMULATION_CONFIG),
    },
    ConfigType.CONTROLLER_CONFIG: DEFAULT_CONTROLLER_CONFIGS,
    ConfigType.SCENARIOS: [],
}


def get_default_exports_file(simulation_type: SimulationType, analysis_type: AnalysisType):
    """Return the default exports file

    Parameters
    ----------
    simulation_type: SimulationType
    analysis_type: AnalysisType

    Returns
    -------
    str

    """
    if simulation_type == SimulationType.SNAPSHOT:
        filename = "snapshot-exports.toml"
    elif simulation_type in (SimulationType.QSTS, SimulationType.TIME_SERIES):
        if analysis_type == AnalysisType.COST_BENEFIT:
            filename = "cba-exports.toml"
        else:
            filename = "exports.toml"
    else:
        assert False, f"Exports for {simulation_type} is not supported."

    return os.path.join(
        os.path.dirname(getattr(disco, "__path__")[0]),
        "disco",
        "pydss",
        "config",
        filename
    )


def get_default_reports_file(simulation_type: SimulationType):
    """Return the default report file for the simulation type.

    Parameters
    ----------
    simulation_type : SimulationType

    Returns
    -------
    str

    """
    if simulation_type in (SimulationType.QSTS, SimulationType.TIME_SERIES):
        filename = "time_series_reports.toml"
    elif simulation_type == SimulationType.SNAPSHOT:
        filename = "snapshot_reports.toml"
    else:
        assert False, simulation_type

    return os.path.join(
        os.path.dirname(getattr(disco, "__path__")[0]),
        "disco",
        "extensions",
        "pydss_simulation",
        filename,
    )


class PyDssConfigurationBase(DistributionConfiguration):
    """Represents the configuration options for a PyDSS simulation."""

    def __init__(self,
                 #add_pmpp_if_missing, # TODO
                 **kwargs):
        """Constructs PyDssConfiguration."""
        super(PyDssConfigurationBase, self).__init__(**kwargs)

        # Kinda hacky, but this enables PyDssConfiguration.deserialize().
        if "pydss_inputs" in kwargs:
            cfg = kwargs["pydss_inputs"]
            self._pydss_inputs = self.deserialize_pydss_inputs(cfg)
        else:
            self._pydss_inputs = self.get_default_pydss_config()

        self._ensure_pydss_controller_registry()

    @property
    def pydss_inputs(self):
        """Get the PyDSS inputs configuration.

        Returns
        -------
        dict

        """
        return copy.deepcopy(self._pydss_inputs)

    def _get_config(self, config_type):
        config = self._pydss_inputs.get(config_type)
        if config is None:
            raise InvalidParameter(f"invalid config_type={config_type}")

        return config

    @staticmethod
    def _ensure_pydss_controller_registry():
        """Ensure DISCO's controllers registered in PyDSS."""
        registry = Registry()
        for item in DEFAULT_CONTROLLER_CONFIGS:
            registered = registry.is_controller_registered(
                controller_type=item["controller_type"],
                name=item["name"]
            )
            if registered:
                continue
            registry.register_controller(
                controller_type=item["controller_type"],
                controller={
                    "name": item["name"],
                    "filename": os.path.abspath(item["filename"])
                }
            )

    def _serialize(self, data):
        data["pydss_inputs"] = self.serialize_pydss_inputs(self._pydss_inputs)

    def get_pydss_config(self, config_type):
        """Return the configuration for the given type.

        Parameters
        ----------
        config_type : ConfigType

        Returns
        -------
        dict

        """
        return copy.deepcopy(self._get_config(config_type))

    def set_pydss_config(self, config_type, config):
        """Set the configuration for the given type.

        Parameters
        ----------
        config_type : ConfigType

        Returns
        -------
        dict

        """
        # This will validate config_type.
        self._get_config(config_type)

        self._pydss_inputs[config_type] = copy.deepcopy(config)
        logger.debug("Set new PyDSS config for %s: %s", config_type, config)

    def add_pydss_controller_config(self, controller_type, name):
        """Add a PyDSS controller to the configuration.

        Parameters
        ----------
        controller_type : str
            The PyDSS controller type
        name : str
            The PyDSS controller name
        """
        controllers = self._pydss_inputs[ConfigType.CONTROLLER_CONFIG]
        for controller in controllers:
            if controller["controller_type"] != controller_type:
                continue
            if controller["name"] == name:
                raise InvalidParameter(f"controller {name} is already stored")

        controller = {
            "controller_type": controller_type,
            "name": name
        }
        self._pydss_inputs[ConfigType.CONTROLLER_CONFIG].append(controller)
        logger.info("Added controller %s", name)

    def remove_pydss_controller_config(self, controller_type, name):
        """Remove a PyDSS controller from the configuration.

        Parameters
        ----------
        controller_type: str
            The PyDSS controller type
        name : str
            The PyDSS controller name

        Raises
        ------
        InvalidParameter
            Raised if name is not stored.

        """
        controllers = self._pydss_inputs[ConfigType.CONTROLLER_CONFIG]
        controller_index = None
        for i, controller in enumerate(controllers):
            if controller["controller_type"] != controller_type:
                continue
            if controller["name"] == name:
                controller_index = i

        if not controller_index:
            raise InvalidParameter(f"{name} is not stored")

        self._pydss_inputs[ConfigType.CONTROLLER_CONFIG].pop(controller_index)
        logger.info("Removed controller %s", name)

    def add_pydss_simulation_config(self, name, simulation):
        """Add a PyDSS simulation config to the configuration.

        Call PyDssConfigurationBase.get_default_pydss_simulation_config() for a
        base simulation.

        Parameters
        ----------
        name : str
        simulation : dict

        """
        self._pydss_inputs[ConfigType.SIMULATION_CONFIG][name] = simulation
        logger.info("Added simulation %s", name)

    def remove_pydss_simulation_config(self, name):
        """Remove a PyDSS simulation config from the configuration.

        Parameters
        ----------
        name : str

        Raises
        ------
        InvalidParameter
            Raised if name is not stored.

        """
        if name not in self._pydss_inputs[ConfigType.SIMULATION_CONFIG]:
            raise InvalidParameter(f"{name} is not stored")

        self._pydss_inputs[ConfigType.SIMULATION_CONFIG].pop(name)
        logger.info("Removed simulation %s", name)

    def get_pydss_simulation_config(self, name):
        """Return the simulation config.

        Parameters
        ----------
        name : str

        Returns
        -------
        dict

        Raises
        ------
        InvalidParameter
            Raised if name is not stored.

        """
        if name not in self._pydss_inputs[ConfigType.SIMULATION_CONFIG]:
            raise InvalidParameter(f"{name} is not stored")

        return self._pydss_inputs[ConfigType.SIMULATION_CONFIG][name]

    @staticmethod
    def serialize_pydss_inputs(inputs):
        """Serializes PyDSS input configuration, replacing enums with values.

        Parameters
        ----------
        inputs : dict

        Returns
        -------
        dict

        """
        return {k.value: v for k, v in inputs.items()}

    @staticmethod
    def deserialize_pydss_inputs(cfg):
        """Deserializes PyDSS input configuration, replacing values with enums.

        Parameters
        ----------
        inputs : dict

        Returns
        -------
        dict

        """
        return {get_enum_from_value(ConfigType, k): v for k, v in cfg.items()}

    @staticmethod
    def get_default_pydss_config():
        """Return the default PyDSS configuration.

        Returns
        -------
        dict

        """
        config = {k: copy.deepcopy(DEFAULT_PYDSS_CONFIG[k]) for k in ConfigType}
        return config

    @staticmethod
    def get_default_pydss_simulation_config():
        """Return the default PyDSS simulation configuration.

        Returns
        -------
        dict

        """
        return copy.deepcopy(DEFAULT_PYDSS_SIMULATION_CONFIG)

    @staticmethod
    def make_default_pydss_scenario(scenario_name, exports=None, post_process_infos=None):
        """Return a default scenario dictionary for a given name.

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        dict

        """
        return {
            "name": scenario_name,
            "exports": {} if exports is None else exports,
            "post_process_infos": [] if post_process_infos is None else post_process_infos,
        }



#!/usr/bin/env python

"""Creates JADE configuration for stage 1 of pydss_simulation pipeline."""

import logging
import sys
from datetime import timedelta

import click

from jade.common import CONFIG_FILE
from jade.loggers import setup_logging
from jade.utils.utils import load_data
from PyDSS.common import SnapshotTimePointSelectionMode
from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO

from disco.enums import SimulationType, AnalysisType
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.pydss.common import ConfigType
from disco.pydss.pydss_configuration_base import (
    get_default_exports_file,
    get_default_reports_file,
    DEFAULT_LOAD_SHAPE_START_TIME,
)
from disco.pydss.common import SCENARIO_NAME_DELIMITER

ESTIMATED_EXEC_SECS_PER_JOB = 10

logger = logging.getLogger(__name__)

COMMON_SNAPSHOT_OPTIONS = (
    click.option(
        "-c",
        "--config-file",
        default=CONFIG_FILE,
        show_default=True,
        help="JADE config file to create",
    ),
    click.option(
        "-e",
        "--exports-filename",
        default=get_default_exports_file(SimulationType.SNAPSHOT, AnalysisType.IMPACT_ANALYSIS),
        show_default=True,
        help="PyDSS export options",
    ),
    click.option(
        "-r",
        "--reports-filename",
        default=get_default_reports_file(SimulationType.SNAPSHOT),
        show_default=True,
        help="PyDSS report options.",
    ),
    click.option(
        "--with-loadshape/--no-with-loadshape",
        is_flag=True,
        required=True,
        help="Configure snapshot simulation with loashape profile.",
    ),
    click.option(
        "--auto-select-time-points/--no-auto-select-time-points",
        is_flag=True,
        default=True,
        show_default=True,
        help="Automatically select the time point based on max PV-load ratio. Only applicable if "
        "--with-loadshape.",
    ),
    click.option(
        "-d",
        "--auto-select-time-points-search-duration-days",
        default=365,
        show_default=True,
        help="Search duration in days. Only applicable with --auto-select-time-points.",
    ),
    click.option(
        "--shuffle/--no-shuffle",
        is_flag=True,
        default=True,
        show_default=True,
        help="Shuffle order of jobs.",
    ),
    click.option(
        "--store-per-element-data/--no-store-per-element-data",
        is_flag=True,
        default=False,
        show_default=True,
        help="Store per-element data in thermal and voltage metrics.",
    ),
    click.option(
        "--strip-whitespace/--no-strip-whitespace",
        is_flag=True,
        default=False,
        show_default=True,
        help="Strip whitespace in file.",
    ),
    click.option(
        "-v",
        "--volt-var-curve",
        default=None,
        help="Update the PyDSS volt-var curve name. If not set, use the pre-configured curve.",
    ),
    click.option(
        "--verbose",
        is_flag=True,
        default=False,
        help="Enable debug logging",
    ),
)


def common_snapshot_options(func):
    for option in reversed(COMMON_SNAPSHOT_OPTIONS):
        func = option(func)
    return func


@click.command()
@click.argument("inputs")
@common_snapshot_options
@click.option(
    "--dc-ac-ratio",
    default=None,
    type=float,
    help="Set a custom DC-AC ratio for PV Systems.",
)
@click.option(
    "--pf1/--no-pf1",
    is_flag=True,
    default=True,
    show_default=True,
    help="Include PF1 scenario or not",
)
@click.option(
    "--control-mode/--no-control-mode",
    is_flag=True,
    default=True,
    show_default=True,
    help="Include control_mode scenario or not",
)
@click.option(
    "--order-by-penetration/--no-order-by-penetration",
    default=False,
    show_default=True,
    help="Make jobs with higher penetration levels blocked by those with lower levels.",
)
def snapshot(
    inputs,
    config_file,
    exports_filename,
    reports_filename,
    with_loadshape,
    auto_select_time_points,
    auto_select_time_points_search_duration_days,
    shuffle,
    store_per_element_data,
    strip_whitespace,
    volt_var_curve,
    verbose,
    dc_ac_ratio,
    pf1,
    control_mode,
    order_by_penetration,
):
    """Create JADE configuration for snapshot simulations."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, None, console_level=level, packages=["disco"])

    if not pf1 and not control_mode:
        logger.error("At least one of '--pf1' or '--control-mode' must be set.")
        sys.exit(1)

    simulation_config, scenarios = make_simulation_config(
        reports_filename,
        exports_filename,
        pf1,
        control_mode,
        store_per_element_data,
        with_loadshape,
        auto_select_time_points,
        auto_select_time_points_search_duration_days,
    )
    config = PyDssConfiguration.auto_config(
        inputs,
        simulation_config=simulation_config,
        scenarios=scenarios,
        order_by_penetration=order_by_penetration,
        estimated_exec_secs_per_job=ESTIMATED_EXEC_SECS_PER_JOB,
        dc_ac_ratio=dc_ac_ratio,
    )

    has_pydss_controllers = config.has_pydss_controllers()
    if control_mode and not has_pydss_controllers:
        scenarios_config = config.get_pydss_config(ConfigType.SCENARIOS)
        indices_to_pop = []
        for i, scenario in enumerate(scenarios_config):
            if CONTROL_MODE_SCENARIO in scenario["name"]:
                indices_to_pop.append(i)
        for i in indices_to_pop.reverse():
            scenarios_config.pop(i)
        logger.info(
            "Excluding %s scenario because there are no pydss controllers.", CONTROL_MODE_SCENARIO
        )
        config.set_pydss_config(ConfigType.SCENARIOS, scenarios_config)

    if volt_var_curve is not None:
        if has_pydss_controllers and control_mode:
            config.update_volt_var_curve(volt_var_curve)
        else:
            logger.warning(
                "Setting a volt_var_curve has no effect when there is no %s scenario.",
                CONTROL_MODE_SCENARIO,
            )

    # We can't currently predict how long each job will take. If we did, we could set
    # estimated_run_minutes for each job.
    # Shuffle the jobs randomly so that we have a better chance of getting batches with similar
    # runtimes.
    if shuffle:
        config.shuffle_jobs()

    if with_loadshape:
        config = switch_snapshot_to_qsts(config)

    indent = None if strip_whitespace else 2
    config.dump(filename=config_file, indent=indent)
    print(f"Created {config_file} for Snapshot Analysis")


def make_simulation_config(
    reports_filename,
    exports_filename,
    include_pf1,
    include_control_mode,
    store_per_element_data,
    with_loadshape,
    auto_select_time_points,
    auto_select_time_points_search_duration_days,
):
    simulation_config = PyDssConfiguration.get_default_pydss_simulation_config()
    simulation_config["reports"] = load_data(reports_filename)["reports"]
    for report in simulation_config["reports"]["types"]:
        if report["name"] in ("Thermal Metrics", "Voltage Metrics"):
            report["store_per_element_data"] = store_per_element_data

    names = []
    scenarios = []
    if not include_control_mode and not include_pf1:
        logger.error("At least one of 'include_pf1' and 'include_control_mode' must be set.")
        sys.exit(1)
    if include_control_mode:
        names.append(CONTROL_MODE_SCENARIO)
    if include_pf1:
        names.append(PF1_SCENARIO)

    if with_loadshape:
        simulation_config["project"]["simulation_type"] = SimulationType.QSTS.value
        if auto_select_time_points:
            for scenario_name in names:
                for mode in SnapshotTimePointSelectionMode:
                    if mode == SnapshotTimePointSelectionMode.NONE:
                        continue
                    name = f"{scenario_name}{SCENARIO_NAME_DELIMITER}{mode.value}"
                    duration_min = float(auto_select_time_points_search_duration_days) * 24 * 60
                    scenario = PyDssConfiguration.make_default_pydss_scenario(name)
                    scenario["snapshot_time_point_selection_config"] = {
                        "mode": mode.value,
                        "start_time": DEFAULT_LOAD_SHAPE_START_TIME,
                        "search_duration_min": duration_min,
                    }
                    scenarios.append(scenario)
        else:
            scenarios = [PyDssConfiguration.make_default_pydss_scenario(x) for x in names]
    else:
        exports = {} if exports_filename is None else load_data(exports_filename)
        simulation_config["project"]["simulation_type"] = SimulationType.SNAPSHOT.value
        scenarios = [PyDssConfiguration.make_default_pydss_scenario(x) for x in names]
        scenarios = [
            PyDssConfiguration.make_default_pydss_scenario(
                x,
                exports=exports,
            )
            for x in names
        ]

    return simulation_config, scenarios


def switch_snapshot_to_qsts(config):
    """Use QSTS at one time point to perform SNAPSHOT simulation with loadshape profile"""
    for job in config.iter_pydss_simulation_jobs():
        job.model.simulation.simulation_type = SimulationType.QSTS
        job.model.simulation.end_time = job.model.simulation.start_time + timedelta(
            seconds=job.model.simulation.step_resolution
        )
    return config



import ast
import time
from .common_functions import *

from jade.utils.timing_utils import track_timing

from disco import timer_stats_collector
from disco.exceptions import ExceededParallelLinesLimit, ExceededParallelTransformersLimit, UpgradesExternalCatalogMissingObjectDefinition
from disco.models.upgrade_cost_analysis_generic_input_model import _extract_specific_model_properties_, TransformerCatalogModel, LineCatalogModel
from disco.models.upgrade_cost_analysis_generic_output_model import TransformerUpgradesTechnicalResultModel, LineUpgradesTechnicalResultModel

logger = logging.getLogger(__name__)

@track_timing(timer_stats_collector)
def correct_line_violations(line_loading_df, line_design_pu, line_upgrade_options, parallel_lines_limit, **kwargs,):
    """This function determines line upgrades to correct line violations.
    It also updates the opendss model with upgrades.

    Parameters
    ----------
    line_loading_df
    line_design_pu
    line_upgrade_options
    parallel_lines_limit

    Returns
    -------

    """
    equipment_type = "Line"
    line_upgrades_df = pd.DataFrame()
    upgrades_dict = {}
    upgrades_dict_parallel = []
    commands_list = []
    # This finds a line code which provides a specified safety margin to a line above its maximum observed loading.
    # If a line code is not found or if line code is too overrated, one or more parallel lines (num_par_lns-1) are added
    overloaded_loading_df = line_loading_df.loc[line_loading_df["status"] == "overloaded"].copy()
    overloaded_loading_df["required_design_amp"] = overloaded_loading_df["max_amp_loading"] / line_design_pu
    deciding_property_list =  _extract_specific_model_properties_(model_name=LineCatalogModel, field_type_key="deciding_property", field_type_value=True)
    line_upgrade_options["kV"] = line_upgrade_options["kV"].round(2)
    overloaded_loading_df["kV"] = overloaded_loading_df["kV"].round(2)
    line_upgrade_options.set_index(deciding_property_list, inplace=True)
    line_upgrade_options = line_upgrade_options.sort_index(level=0)
    overloaded_loading_df.set_index(deciding_property_list, inplace=True)  
    oversize_limit = 2  # limit to determine if chosen upgrade option is too oversized
    extreme_loading_threshold = 2.25  # from observations, if equipment loading is greater than this factor, it is considered extremely overloaded
    # in such extremely loaded cases, the equipment is oversized much more, to avoid iterations in upgrades    
    if len(overloaded_loading_df) > 0:  # if overloading exists
        # iterate over each overloaded line to find a solution
        for index, row in overloaded_loading_df.iterrows():
            logger.debug(row["name"])
            try:
                options = line_upgrade_options.loc[index]
            except KeyError:
                raise UpgradesExternalCatalogMissingObjectDefinition(f"Line of type {dict(zip(deciding_property_list, list(index)))} not found in catalog."
                                                                     f" Please ensure catalog is complete.")
            if isinstance(options, pd.Series):  # if only one option is present, it is returned as series
                options = pd.DataFrame([options])  # convert to required DataFrame format
                options = options.rename_axis(deciding_property_list)  # assign names to the index
            options = options.reset_index().sort_values("normamps")
            if row["max_per_unit_loading"] > extreme_loading_threshold:  # i.e. equipment is very overloaded, then oversize more to avoid multiple upgrade iterations
                row["required_design_amp"] = row["required_design_amp"] * (row["max_per_unit_loading"] * 0.5)
            chosen_option = options.loc[options["normamps"] >= row["required_design_amp"]].sort_values("normamps")
            # TODO: TOGGLE FLAG - EXPLORE IF WE CAN HAVE OPTIONS FOR: PARALLEL, REPLACE,
            # if one chosen option exists and is not very oversized (which is determined by acceptable oversize limit)
            # edit existing line and change line configuration/ampacity
            if (len(chosen_option) != 0) and \
                    (chosen_option["normamps"] <= oversize_limit * row["required_design_amp"]).any():
                chosen_option = chosen_option.sort_values("normamps")
                chosen_option = chosen_option.iloc[0]  # choose lowest available option
                new_config_type = chosen_option["line_definition_type"]     
                new_config_name = chosen_option[new_config_type].lower()
                temp_commands_list = []
                # edit existing line
                if (new_config_type == "geometry") or (new_config_type == "linecode"):
                    external_upgrades_technical_catalog = kwargs.get("external_upgrades_technical_catalog", None)
                    command_string = ensure_line_config_exists(chosen_option, new_config_type, external_upgrades_technical_catalog)
                    if command_string is not None:  # if new line config definition had to be added
                        temp_commands_list.append(command_string)                                                    
                    command_string = f"Edit Line.{row['name']} {new_config_type}={new_config_name} normamps={chosen_option['normamps']}"
                    temp_commands_list.append(command_string)
                # if line geometry and line code is not available
                else:
                    command_string = define_line_object(line_name=row['name'], chosen_option=chosen_option.copy(deep=True), action_type="Edit")
                    temp_commands_list.append(command_string)
                # create dictionary of original equipment
                upgrades_dict[row["name"]] = {}
                upgrades_dict[row["name"]]["original_equipment"] = row.to_dict()
                upgrades_dict[row["name"]]["original_equipment"].update({"equipment_type": equipment_type,
                                                                         "upgrade_type": "upgrade",
                                                                         # "parameter_type": "original_equipment",
                                                                         "action": "remove"})
                upgrades_dict[row["name"]]["original_equipment"] .update(chosen_option[["Switch", "phases",
                                                                                        "line_placement"]].to_dict())
                # create dictionary of new equipment
                upgrades_dict[row["name"]]["new_equipment"] = chosen_option.to_dict()
                # line length units for new_equipment should be same as that of original equipment
                upgrades_dict[row["name"]]["new_equipment"]["units"] = upgrades_dict[row["name"]]["original_equipment"]["units"]  
                upgrades_dict[row["name"]]["new_equipment"].update({"equipment_type": equipment_type,
                                                                    "upgrade_type": "upgrade",
                                                                    # "parameter_type": "new_equipment",
                                                                    "action": "add", "name": row["name"]})
                # run command for upgraded equipment, that resolves overloading for one equipment
                for command_item in temp_commands_list:
                    check_dss_run_command(command_item)
                    circuit_solve_and_check(raise_exception=True, **kwargs)
                commands_list = commands_list + temp_commands_list
            # if higher upgrade is not available or chosen line upgrade rating is much higher than required,
            # dont oversize. Instead, place lines in parallelma
            else:
                external_upgrades_technical_catalog = kwargs.get("external_upgrades_technical_catalog", None)
                parallel_line_commands, temp_upgrades_dict_parallel = identify_parallel_lines(options=options, object_row=row,
                                                                                         parallel_lines_limit=parallel_lines_limit, 
                                                                                         external_upgrades_technical_catalog=external_upgrades_technical_catalog)
                # run command for all parallel equipment added, that resolves overloading for one equipment
                for command_item in parallel_line_commands:
                    check_dss_run_command(command_item)
                    check_dss_run_command('CalcVoltageBases')
                    circuit_solve_and_check(raise_exception=True, **kwargs)
                commands_list = commands_list + parallel_line_commands
                upgrades_dict_parallel = upgrades_dict_parallel + temp_upgrades_dict_parallel  # parallel upgrades is stored in a list (since it has same original_equipment name)
        index_names = ["original_equipment_name", "parameter_type"]
        if upgrades_dict:  # if dictionary is not empty
            line_upgrades_df = create_dataframe_from_nested_dict(user_dict=upgrades_dict, index_names=index_names)
        line_upgrades_df = pd.concat([line_upgrades_df, pd.DataFrame(upgrades_dict_parallel)])
        line_upgrades_df.rename(columns={"name": "final_equipment_name"}, inplace=True)
        line_upgrades_df = line_upgrades_df.set_index(
            ["equipment_type", "upgrade_type", "parameter_type", "action", "final_equipment_name",
             "original_equipment_name"]).reset_index()
    else:  # if there is no overloading
        logger.info("This case has no line violations")
    circuit_solve_and_check(raise_exception=True, **kwargs)  # this is added as a final check for convergence
    output_fields =  list(LineUpgradesTechnicalResultModel.schema(True).get("properties").keys())  # get fields with alias
    line_upgrades_df = line_upgrades_df[output_fields]
    return commands_list, line_upgrades_df


@track_timing(timer_stats_collector)
def identify_parallel_lines(options, object_row, parallel_lines_limit, **kwargs):
    """This function identifies parallel line solutions, when a direct upgrade solution is not available from catalogue

    Parameters
    ----------
    options
    row
    parallel_lines_limit

    Returns
    -------

    """
    output_fields = list(LineUpgradesTechnicalResultModel.schema(True).get("properties").keys())
    temp_output_fields = set(output_fields) - {"final_equipment_name"}
    commands_list = []
    temp_dict = {}
    # calculate number of parallel lines needed to carry remaining amperes (in addition to existing line)
    options["num_parallel_raw"] = (object_row["required_design_amp"] - object_row["normamps"]) / options["normamps"]
    options["num_parallel"] = options["num_parallel_raw"].apply(np.ceil)
    options["choose_parallel_metric"] = options["num_parallel"] - options["num_parallel_raw"]
    # choose option that has the least value of this metric- since that represents the per unit oversizing
    chosen_option = pd.DataFrame(options.loc[options["choose_parallel_metric"].idxmin()]).T
    chosen_option = chosen_option.sort_values("normamps")
    chosen_option = chosen_option.iloc[0]  # choose lowest available option
    num_parallel_lines = int(chosen_option["num_parallel"])
    if num_parallel_lines > parallel_lines_limit:
        raise ExceededParallelLinesLimit(f"Number of parallel lines required is {num_parallel_lines}, which exceeds limit of {parallel_lines_limit}."
                                          f" Increase parameter parallel_lines_limit in input config or ensure higher sized equipment is present in technical catalog.")
    new_config_type = chosen_option["line_definition_type"]
    external_upgrades_technical_catalog = kwargs.get("external_upgrades_technical_catalog", None)
    upgrades_dict_parallel = []
    for i in range(0, num_parallel_lines):
        curr_time = str(time.time())
        unique_tag = curr_time.split(".")[0] + "_" + curr_time.split(".")[1] + "_" + str(i)
        new_name = "upgrade_" + object_row["name"] + unique_tag
        chosen_option["name"] = new_name
        temp_dict = {}
        temp_dict.update(chosen_option.to_dict())
        temp_dict["length"] = object_row["length"]
        temp_dict["units"] = object_row["units"]
        temp_dict.update({"equipment_type": "Line",
                          "original_equipment_name": object_row["name"],
                          "upgrade_type": "new_parallel",
                          "parameter_type": "new_equipment",
                          "action": "add"})
        # # if too much info, can remove this metrics from output. For now, these are included.
        # remove_fields = ['num_parallel_raw', 'num_parallel', 'choose_parallel_metric']
        # for x in remove_fields:
        #     temp_dict.pop(x)
        if (new_config_type == "geometry") or (new_config_type == "linecode"):
            config_command_string = ensure_line_config_exists(chosen_option, new_config_type, external_upgrades_technical_catalog)
            if config_command_string is not None:  # if new line config definition had to be added
                commands_list.append(config_command_string)          
            command_string = f"New Line.{new_name} bus1={object_row['bus1']} bus2={object_row['bus2']} length={object_row['length']} " \
                    f"units={object_row['units']} {new_config_type}={chosen_option[new_config_type]} " \
                    f"phases={chosen_option['phases']} normamps={chosen_option['normamps']} enabled=Yes"
            commands_list.append(command_string)
        # if line geometry and line code is not available, add impedance properties directly
        else:
            command_string = define_line_object(line_name=new_name, chosen_option=chosen_option.copy(deep=True), action_type="New", 
                               bus1=object_row['bus1'], bus2=object_row['bus2'], length=object_row['length'],
                               original_units=object_row['units'])
            commands_list.append(command_string)
        temp_dict.pop("choose_parallel_metric")
        temp_dict.pop("num_parallel_raw")
        missing_fields = list(set(temp_output_fields).difference(temp_dict.keys()))  # find missing fields in temp_dict
        temp_dict.update(object_row[missing_fields].to_dict())
        upgrades_dict_parallel.append(temp_dict)
    return commands_list, upgrades_dict_parallel


def define_line_object(line_name, chosen_option, action_type, **kwargs):
    """This function is used to create a command string to define an opendss line object
    """
    command_string = f"{action_type} Line.{line_name} normamps={chosen_option['normamps']} emergamps={chosen_option['emergamps']}"
    if action_type == "New":
        bus1 = kwargs.get("bus1", None)
        bus2 = kwargs.get("bus2", None)
        length = kwargs.get("length", None)
        original_units = kwargs.get("original_units", None)
        if bus1 is None or bus2 is None or length is None or original_units is None:
            raise ValueError(f"Bus and length information is needed when defining a new line object.")
        if original_units != chosen_option["units"]:  # if units are different
            # keep units of chosen option because impedances are defined in this unit
            units = chosen_option["units"]  
            # change length to be reflected in the new units
            length = convert_length_units(length, unit_in=original_units, unit_out=chosen_option["units"])
        else:
            units = original_units
            
        command_string = command_string + f" bus1={bus1} bus2={bus2} Length={length} " \
                         f"Units={units} phases={chosen_option['phases']}"
    # Impedances may be specified by symmetrical component values or by matrix values
    # (refer OpenDSS manual for more information on these parameters)
    matrix_impedance_prop = _extract_specific_model_properties_(model_name=LineCatalogModel, field_type_key="matrix_impedance_property", field_type_value=True)
    symm_impedance_prop = _extract_specific_model_properties_(model_name=LineCatalogModel, field_type_key="symmetrical_impedance_property", field_type_value=True)
    
    method = "matrix_impedance"  # first try matrix impedance, else add symm component
    matrix_fields = [s for s in matrix_impedance_prop if 'matrix' in s]
    for field in matrix_fields:
        chosen_option[field] = str(chosen_option[field]).replace("'","")
        chosen_option[field] = chosen_option[field].replace("[","(")
        chosen_option[field] = chosen_option[field].replace("]",")")
    for property_name in matrix_impedance_prop:
        if chosen_option[property_name]:
            temp_s = f" {property_name}={chosen_option[property_name]}"
            command_string = command_string + temp_s
        else:
            method = "symmetrical_impedance"
            break       
    if method == "symmetrical_impedance":
        for property_name in symm_impedance_prop:
            if chosen_option[property_name]:
                temp_s = f" {property_name}={chosen_option[property_name]}"
                command_string = command_string + temp_s
            else:
                break
    command_string = command_string + " enabled=Yes"
    return command_string


def define_xfmr_object(xfmr_name, xfmr_info_series, action_type, buses_list=None):
    """This function is used to create a command string to define an opendss transformer object

    Parameters
    ----------
    xfmr_name
    xfmr_info_series
    action_type
    buses_list

    Returns
    -------
    string
    """
    if isinstance(xfmr_info_series["conns"], str):
        xfmr_info_series["conns"] = ast.literal_eval(xfmr_info_series["conns"])
    if isinstance(xfmr_info_series["kVs"], str):
        xfmr_info_series["kVs"] = ast.literal_eval(xfmr_info_series["kVs"])
    if isinstance(xfmr_info_series["%Rs"], str):
        xfmr_info_series["%Rs"] = ast.literal_eval(xfmr_info_series["%Rs"])
    if isinstance(xfmr_info_series["kVAs"], str):
        xfmr_info_series["kVAs"] = ast.literal_eval(xfmr_info_series["kVAs"])
    command_string = f"{action_type} Transformer.{xfmr_name}"
    if action_type == "New":
        if buses_list is None:
            raise Exception("Buses list has to be passed if defining new transformer object.")
        s_temp = " buses=("
        for bus in buses_list:
            s_temp = s_temp + f"{bus} "
        s_temp = s_temp + ")"
        s_temp = s_temp.replace(" )", ")")
        command_string = command_string + s_temp + f" phases={xfmr_info_series['phases']} " \
                                                   f"windings={xfmr_info_series['windings']}"
    for wdg_count in range(xfmr_info_series["wdg"]):
        temp_s = f" wdg={wdg_count+1} kVA={xfmr_info_series['kVAs'][wdg_count]} kV={xfmr_info_series['kVs'][wdg_count]} " \
                 f"conn={xfmr_info_series['conns'][wdg_count]} %r={xfmr_info_series['%Rs'][wdg_count]} " \
                 f"rneut={xfmr_info_series['Rneut']} xneut={xfmr_info_series['Xneut']}"
        command_string = command_string + temp_s
    if xfmr_info_series["wdg"] > 1:
        # define reactances between windings
        temp_s = f" XLT={xfmr_info_series['XLT']} XHT={xfmr_info_series['XLT']} XHL={xfmr_info_series['XHL']}"
        command_string = command_string + temp_s
    if xfmr_info_series["phases"] > 1:
        # for higher phase order transformers, also define xscarray
        temp_s = f" Xscarray={xfmr_info_series['Xscarray']}"
        command_string = command_string + temp_s
    # these properties contain general transformer rating data
    # (refer OpenDSS manual for more information on these parameters)
    general_property_list =  _extract_specific_model_properties_(model_name=TransformerCatalogModel, field_type_key="write_property", field_type_value=True)
    for property_name in general_property_list:
        temp_s = f" {property_name}={xfmr_info_series[property_name]}"
        command_string = command_string + temp_s
    command_string = command_string + " enabled=Yes"
    return command_string


@track_timing(timer_stats_collector)
def correct_xfmr_violations(xfmr_loading_df, xfmr_design_pu, xfmr_upgrade_options,
                            parallel_transformers_limit, **kwargs):
    """This function determines transformer upgrades to correct transformer violations.
    It also updates the opendss model with upgrades.

    Parameters
    ----------
    xfmr_loading_df
    xfmr_design_pu
    xfmr_upgrade_options
    parallel_transformers_limit

    Returns
    -------

    """
    equipment_type = "Transformer"
    xfmr_upgrades_df = pd.DataFrame()
    upgrades_dict = {}
    upgrades_dict_parallel = []
    commands_list = []
    # This finds a line code which provides a specified safety margin to a line above its maximum observed loading.
    # If a line code is not found or if line code is too overrated, one or more parallel lines (num_par_lns-1) are added
    overloaded_loading_df = xfmr_loading_df.loc[xfmr_loading_df["status"] == "overloaded"].copy()
    overloaded_loading_df["required_design_amp"] = overloaded_loading_df["max_amp_loading"] / xfmr_design_pu
    # list of properties based on which upgrade is chosen
    deciding_property_list =  _extract_specific_model_properties_(model_name=TransformerCatalogModel, field_type_key="deciding_property", field_type_value=True)
    # round to same precision for upgrade options, and overloaded equipment
    xfmr_upgrade_options["kV"] = xfmr_upgrade_options["kV"].round(2)
    overloaded_loading_df["kV"] = overloaded_loading_df["kV"].round(2)
    xfmr_upgrade_options["kVs"] = xfmr_upgrade_options["kVs"].apply(lambda x: [round(a, 2) for a in x])
    overloaded_loading_df["kVs"] = overloaded_loading_df["kVs"].apply(lambda x: [round(a, 2) for a in x])
    # convert lists to string type (so they can be set as dataframe index later)
    xfmr_upgrade_options[['conns', 'kVs']] = xfmr_upgrade_options[['conns', 'kVs']].astype(str)
    overloaded_loading_df[['conns', 'kVs']] = overloaded_loading_df[['conns', 'kVs']].astype(str)
    xfmr_upgrade_options.set_index(deciding_property_list, inplace=True)
    xfmr_upgrade_options = xfmr_upgrade_options.sort_index(level=0)
    overloaded_loading_df.set_index(deciding_property_list, inplace=True)
    equipment_oversize_limit = 2  # limit to determine if chosen upgrade option is too oversized
    extreme_loading_threshold = 2.25  # if equipment loading is greater than this factor, it is considered extremely overloaded
    # in such extremely loaded cases, the equipment is oversized much more, to avoid iterations in upgrades
    if len(overloaded_loading_df) > 0:  # if overloading exists
        xfmr_upgrades_df = pd.DataFrame()
        # iterate over each overloaded line to find a solution
        for index, row in overloaded_loading_df.iterrows():
            try:
                options = xfmr_upgrade_options.loc[index]
            except KeyError:
                raise UpgradesExternalCatalogMissingObjectDefinition(f"Transformer of type {dict(zip(deciding_property_list, list(index)))} not found in catalog."
                                                                     f" Please ensure catalog is complete.")
            if isinstance(options, pd.Series):  # if only one option is present, it is returned as series
                options = pd.DataFrame([options])  # convert to required DataFrame format
                options = options.rename_axis(deciding_property_list)  # assign names to the index
            options = options.reset_index().sort_values("amp_limit_per_phase")
            if row["max_per_unit_loading"] > extreme_loading_threshold:  # i.e. equipment is very overloaded, then oversize to avoid multiple upgrade iterations
                row["required_design_amp"] = row["required_design_amp"] * (row["max_per_unit_loading"] * 0.5)
            chosen_option = options.loc[options["amp_limit_per_phase"] >=
                                        row["required_design_amp"]].sort_values("amp_limit_per_phase")
            # if one chosen option exists and is not very oversized (which is determined by acceptable oversize limit)
            # edit existing object
            if (len(chosen_option) != 0) and \
                    (chosen_option["amp_limit_per_phase"] <= equipment_oversize_limit * row["required_design_amp"]).any():
                chosen_option = chosen_option.sort_values("amp_limit_per_phase")
                chosen_option = chosen_option.iloc[0]  # choose lowest available option
                chosen_option["conns"] = ast.literal_eval(chosen_option["conns"])
                chosen_option["kVs"] = ast.literal_eval(chosen_option["kVs"])
                if isinstance(chosen_option["%Rs"], str):
                    chosen_option["%Rs"] = ast.literal_eval(chosen_option["%Rs"])
                if isinstance(chosen_option["kVAs"], str):
                    chosen_option["kVAs"] = ast.literal_eval(chosen_option["kVAs"])
                # edit existing transformer
                command_string = define_xfmr_object(xfmr_name=row["name"], xfmr_info_series=chosen_option,
                                                    action_type="Edit")
                commands_list.append(command_string)
                # create dictionary of original equipment
                upgrades_dict[row["name"]] = {}
                upgrades_dict[row["name"]]["original_equipment"] = row.to_dict()
                upgrades_dict[row["name"]]["original_equipment"].update({"equipment_type": equipment_type,
                                                                         "upgrade_type": "upgrade",
                                                                         # "parameter_type": "original_equipment",
                                                                         "action": "remove"})
                upgrades_dict[row["name"]]["original_equipment"].update(chosen_option[deciding_property_list].to_dict())
                # create dictionary of new equipment
                upgrades_dict[row["name"]]["new_equipment"] = chosen_option.to_dict()
                upgrades_dict[row["name"]]["new_equipment"].update({"equipment_type": equipment_type,
                                                                    "upgrade_type": "upgrade",
                                                                    # "parameter_type": "new_equipment",
                                                                    "action": "add", "name": row["name"]})
                check_dss_run_command(command_string)  # run command for upgraded equipment
                circuit_solve_and_check(raise_exception=True, **kwargs)
            # if higher upgrade is not available or chosen upgrade rating is much higher than required,
            # dont oversize. Instead, place equipment in parallel
            else:
                parallel_xfmr_commands, temp_upgrades_dict_parallel = identify_parallel_xfmrs(upgrade_options=options, object_row=row,
                                                                                         parallel_transformers_limit=parallel_transformers_limit)
                # run command for all new parallel equipment added, that resolves overloading for one equipment
                for command_item in parallel_xfmr_commands:
                    check_dss_run_command(command_item)
                    check_dss_run_command('CalcVoltageBases')
                    circuit_solve_and_check(raise_exception=True, **kwargs)
                commands_list = commands_list + parallel_xfmr_commands
                upgrades_dict_parallel = upgrades_dict_parallel + temp_upgrades_dict_parallel  # parallel upgrades is stored in a list (since it has same original_equipment name)
        index_names = ["original_equipment_name", "parameter_type"]
        if upgrades_dict:  # if dictionary is not empty
            xfmr_upgrades_df = create_dataframe_from_nested_dict(user_dict=upgrades_dict, index_names=index_names)
        xfmr_upgrades_df = pd.concat([xfmr_upgrades_df, pd.DataFrame(upgrades_dict_parallel)])
        xfmr_upgrades_df.rename(columns={"name": "final_equipment_name"}, inplace=True)
        xfmr_upgrades_df = xfmr_upgrades_df.set_index(
            ["equipment_type", "upgrade_type", "parameter_type", "action", "final_equipment_name",
             "original_equipment_name"]).reset_index()

    else:  # if there is no overloading
        logger.info("This case has no transformer violations")
    circuit_solve_and_check(raise_exception=True, **kwargs)  # this is added as a final check for convergence
    output_fields =  list(TransformerUpgradesTechnicalResultModel.schema(True).get("properties").keys())  # get fields with alias
    xfmr_upgrades_df = xfmr_upgrades_df[output_fields]
    return commands_list, xfmr_upgrades_df


@track_timing(timer_stats_collector)
def identify_parallel_xfmrs(upgrade_options, object_row, parallel_transformers_limit):
    """This function identifies parallel transformer solutions, when a direct upgrade solution is not available from catalogue

    Parameters
    ----------
    options
    row
    parallel_transformers_limit

    Returns
    -------

    """
    output_fields = list(TransformerUpgradesTechnicalResultModel.schema(True).get("properties").keys())
    temp_output_fields = set(output_fields) - {"final_equipment_name"}
    equipment_type = "Transformer"
    commands_list = []
    upgrades_dict_parallel = {}
    # calculate number of parallel equipment needed to carry remaining amperes (in addition to existing equipment)
    upgrade_options["num_parallel_raw"] = (object_row["required_design_amp"] -
                                   object_row["amp_limit_per_phase"]) / upgrade_options["amp_limit_per_phase"]
    upgrade_options["num_parallel"] = upgrade_options["num_parallel_raw"].apply(np.ceil)
    upgrade_options["choose_parallel_metric"] = upgrade_options["num_parallel"] - upgrade_options["num_parallel_raw"]
    n = upgrade_options['num_parallel'].min()
    upgrade_options = upgrade_options.loc[upgrade_options["num_parallel"] <= parallel_transformers_limit]
    if len(upgrade_options) == 0:
        raise ExceededParallelTransformersLimit(f"Number of parallel transformers required is {int(n)}, which exceeds limit of {parallel_transformers_limit}."
                                                f" Increase parameter parallel_transformers_limit in input config or ensure higher sized equipment is present in technical catalog.")
    # choose option that has the least value of this metric- since that represents the per unit oversizing
    chosen_option = upgrade_options.loc[upgrade_options["choose_parallel_metric"].idxmin()].copy()
    chosen_option["conns"] = ast.literal_eval(chosen_option["conns"])
    chosen_option["kVs"] = ast.literal_eval(chosen_option["kVs"])
    if isinstance(chosen_option["%Rs"], str):
        chosen_option["%Rs"] = ast.literal_eval(chosen_option["%Rs"])
    if isinstance(chosen_option["kVAs"], str):
        chosen_option["kVAs"] = ast.literal_eval(chosen_option["kVAs"])
    num_parallel_xfmrs = int(chosen_option["num_parallel"])
    upgrades_dict_parallel = []
    for i in range(0, num_parallel_xfmrs):
        curr_time = str(time.time())
        unique_tag = curr_time.split(".")[0] + "_" + curr_time.split(".")[1] + "_" + str(i)
        new_name = "upgrade_" + object_row["name"] + unique_tag
        chosen_option["name"] = new_name
        temp_dict = {}
        temp_dict.update(chosen_option.to_dict())
        temp_dict.update({"equipment_type": equipment_type,
                        "original_equipment_name": object_row["name"],
                        "upgrade_type": "new_parallel",
                        "parameter_type": "new_equipment",
                        "action": "add"})
        command_string = define_xfmr_object(xfmr_name=new_name, xfmr_info_series=chosen_option, action_type="New",
                                            buses_list=object_row["buses"])
        commands_list.append(command_string)
        temp_dict.pop("choose_parallel_metric")
        temp_dict.pop("num_parallel_raw")
        missing_fields = list(set(temp_output_fields).difference(temp_dict.keys()))  # find missing fields in temp_dict
        temp_dict.update(object_row[missing_fields].to_dict())
        upgrades_dict_parallel.append(temp_dict)
    return commands_list, upgrades_dict_parallel



"""Implements functionality for QSTS inputs."""


from collections import namedtuple
import enum
import logging
import os
import re

from jade.jobs.job_inputs_interface import JobInputsInterface
from jade.exceptions import InvalidParameter
from jade.utils.utils import handle_file_not_found, handle_key_error, load_data
from disco.enums import get_placement_from_value
from disco.sources.base import DEFAULT_PV_DEPLOYMENTS_DIRNAME


logger = logging.getLogger(__name__)


class SourceTree1ModelInputs(JobInputsInterface):
    """Implements functionality for Source Type 1 inputs."""

    _DEPLOYMENT_FILENAME = "PVSystems.dss"
    _LOAD_SHAPES_FILENAME = "LoadShapes.dss"
    _KEYS_TYPE = namedtuple(
        "SourceTree1Keys", "substation, feeder, placement"
    )
    _PV_CONFIG_FILENAME = "pv_config.json"
    _METADATA_DIRNAME = "metadata"
    SUBSTATION_DELIMITER = "--"

    def __init__(self, base_directory, pv_deployments_dirname=None):
        """Constructs SourceTree1ModelInputs.

        Parameters
        ----------
        base_directory : str

        """
        self._base = base_directory
        self._pv_deployments_dirname = pv_deployments_dirname or DEFAULT_PV_DEPLOYMENTS_DIRNAME

        if not os.path.exists(self._base):
            raise InvalidParameter("inputs directory does not exist: {}"
                                   .format(self._base))

        self._parameters = self._parse_directories()
        logger.debug("Created %s at %s", self.__class__.__name__, self._base)

    @property
    def base_directory(self):
        return self._base

    def create_key(self, substation, feeder, placement):
        """Create a parameter key for accessing samples and penetration
        levels.

        Parameters
        ----------
        substation : str
        feeder : str
        placement : Placement

        Returns
        -------
        namedtuple

        """
        return self._KEYS_TYPE._make((substation, feeder, placement))

    def get_available_parameters(self):
        """Return a dictionary containing all available parameters.

        Returns
        -------
        dict

            dictionary layout::

                {
                    (Substation, Feeder, Placement) : {
                        samples: Set(
                            penetration_levels,
                        )
                    }
                }

        """
        return self._parameters

    def get_metadata_directory(self, substation, feeder):
        """Get the path of feeder metadata directory"""
        return os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._METADATA_DIRNAME
        )

    def get_deployment_file(self, key, sample, penetration_level):
        """Get the path to an input deployment file for the given parameters.

        Parameters
        ----------
        key : namedtuple
        sample : int
        penetration_level : int

        Returns
        -------
        str
            full path to file

        """
        filename = os.path.join(
            self._base,
            key.substation,
            self._feeder_dirname(key.substation, key.feeder),
            self._pv_deployments_dirname,
            key.placement.value,
            str(sample),
            str(penetration_level),
            self._DEPLOYMENT_FILENAME,
        )

        if not os.path.exists(filename):
            raise InvalidParameter(f"file does not exist: {filename}")
        return filename

    def get_dss_filename(self, substation, feeder, filename):
        """Return the path to DSS file.

        Parameters
        ----------
        substation : str
        feeder : str
        filename : str

        Returns
        -------
        str
            path to file

        Raises
        ------
        InvalidParameter
            Raised if the file does not exist.

        """
        dss_filename = os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            filename,
        )

        if not os.path.exists(dss_filename):
            raise InvalidParameter(f"file does not exist: {dss_filename}")

        return dss_filename

    def get_loadshapes_filename(self, substation, feeder):
        """Return the path to the load shapes file.

        Parameters
        ----------
        feeder : str

        Returns
        -------
        str

        """
        return os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._LOAD_SHAPES_FILENAME,
        )

    def get_substation_opendss_directory(self, substation):
        """Return the path to the OpenDSS files.

        Parameters
        ----------
        substation : str

        Returns
        -------
        str

        """
        return os.path.join(
            self._base,
            substation,
        )

    def get_opendss_directory(self, substation, feeder):
        """Return the path to the OpenDSS files.

        Parameters
        ----------
        substation : str
        feeder : str

        Returns
        -------
        str

        """
        return os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
        )

    def list_substations(self):
        """List available substations.

        Returns
        -------
        list of str
            available substations

        """
        return self._list_available_params("substation")

    def list_feeders(self, substation):
        """List available feeders.

        Parameters
        ----------
        substation : str

        Returns
        -------
        list of str
            available feeders

        """
        return self._list_feeders(substation)

    def list_placements(self):
        """List available placements.

        Returns
        -------
        list of Placement
            available placements

        """
        return self._list_available_params("placement")

    @handle_key_error
    def list_samples(self, key):
        """List available samples.

        Parameters
        ----------
        key : namedtuple

        Returns
        -------
        list of int
            available samples

        """
        samples = list(self._parameters[key].keys())
        samples.sort()
        return samples

    @handle_key_error
    def list_penetration_levels(self, key, sample):
        """List available penetration levels.

        Parameters
        ----------
        key : namedtuple
        sample : int

        Returns
        -------
        list of int
            available penetration levels

        """
        levels = list(self._parameters[key][sample])
        levels.sort()
        return levels

    def list_pv_configs(self, substation, feeder, placement, sample):
        """Return the configurations for each PVSystem.
        Parameters
        ----------
        substation : str
        feeder : str
        placement : Placement
        sample : int
        Returns
        -------
        list
        """
        return load_data(os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._pv_deployments_dirname,
            placement.value,
            str(sample),
            self._PV_CONFIG_FILENAME,
        ))["pv_systems"]

    @staticmethod
    def _get_items_from_directory(path):
        """Strips out miscellaneous files"""
        return [x for x in os.listdir(path)
                if os.path.isdir(os.path.join(path, x))]

    @staticmethod
    def _feeder_dirname(substation, feeder):
        return SourceTree1ModelInputs.SUBSTATION_DELIMITER.join((substation, feeder))

    @staticmethod
    def _feeder_from_dirname(dirname):
        fields = dirname.split(SourceTree1ModelInputs.SUBSTATION_DELIMITER)
        if len(fields) != 2:
            return None
        return fields[1]

    def _list_available_params(self, param):
        params = set()
        is_enum = False
        for key in self._parameters:
            value = key._asdict()[param]
            assert value is not None
            params.add(value)
            if isinstance(value, enum.Enum):
                is_enum = True

        key = None
        if is_enum:
            key = lambda x: x.value
        return sorted(list(params), key=key)

    @handle_file_not_found
    def _list_substations(self):
        return self._get_items_from_directory(self._base)

    @handle_file_not_found
    def _list_feeders(self, substation):
        feeders = []
        feeder_path = os.path.join(self._base, substation)
        for item in self._get_items_from_directory(feeder_path):
            feeder = self._feeder_from_dirname(item)
            if feeder is not None:
                feeders.append(feeder)
        return feeders

    @handle_file_not_found
    def _list_placements(self, substation, feeder):
        placement_path = os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._pv_deployments_dirname,
        )
        if not os.path.exists(placement_path):
            return []

        placements = []
        for x in self._get_items_from_directory(placement_path):
            placement = get_placement_from_value(x)
            placements.append(placement)
        placements.sort(key=lambda x: x.value)
        return placements

    @handle_file_not_found
    def _list_samples(self, substation, feeder, placement):
        sample_path = os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._pv_deployments_dirname,
            placement.value,
        )
        samples = self._get_items_from_directory(sample_path)
        return sorted([int(x) for x in samples])

    @handle_file_not_found
    def _list_penetration_levels(self, substation, feeder, placement, sample):
        penetration_path = os.path.join(
            self._base,
            substation,
            self._feeder_dirname(substation, feeder),
            self._pv_deployments_dirname,
            placement.value,
            str(sample),
        )

        levels = []
        regex = re.compile(r"^(\d+)$")
        for item in os.listdir(penetration_path):
            if not os.path.isdir(os.path.join(penetration_path, item)):
                continue
            match = regex.search(item)
            if match:
                levels.append(int(match.group(1)))

        levels.sort()
        return levels

    def _parse_directories(self):
        data = {}
        for substation in self._list_substations():
            self._parse_feeders(data, substation)

        return data

    def _parse_feeders(self, data, substation):
        for feeder in self._list_feeders(substation):
            self._parse_placements(data, substation, feeder)

    def _parse_placements(self, data, substation, feeder):
        for placement in self._list_placements(substation, feeder):
            self._parse_samples(data, substation, feeder, placement)

    def _parse_samples(self, data, substation, feeder, placement):
        key = self._KEYS_TYPE(substation, feeder, placement)
        data[key] = {}
        for sample in self._list_samples(substation, feeder, placement):
            data[key][sample] = set()
            self._parse_penetrations(data[key][sample], substation, feeder,
                                     placement, sample)

    def _parse_penetrations(self, data, substation, feeder, placement,
                            sample):
        for level in self._list_penetration_levels(substation, feeder, placement, sample):
            data.add(level)



"""Main CLI command for disco."""

import logging
from pathlib import Path

import click

from jade.utils.subprocess_manager import check_run_command

import disco
from disco.cli.config import config

# from disco.cli.configure_analysis import generate_analysis
from disco.cli.simulation_models import simulation_models
from disco.cli.download_source import download_source
from disco.cli.transform_model import transform_model
from disco.cli.prescreen_pv_penetration_levels import prescreen_pv_penetration_levels
from disco.cli.pv_deployments import pv_deployments
from disco.cli.create_pipeline import create_pipeline
from disco.cli.ingest_tables import ingest_tables
from disco.cli.make_summary_tables import make_summary_tables
from disco.cli.compute_hosting_capacity import compute_hosting_capacity
from disco.cli.plots import plot
from disco.cli.summarize_hosting_capacity import summarize_hosting_capacity
from disco.cli.config_generic_models import config_generic_models
from disco.cli.upgrade_cost_analysis import upgrade_cost_analysis
from disco.cli.pydss_hosting_capacity import hosting_capacity_by_timestep


logger = logging.getLogger(__name__)


@click.command()
def install_extensions():
    """Install DISCO's JADE extensions."""
    ext_path = Path(disco.__path__[0]) / "extensions" / "jade_extensions.json"
    check_run_command(f"jade extensions register {ext_path}")


@click.group()
def cli():
    """Entry point"""


# cli.add_command(generate_analysis)
cli.add_command(config)
cli.add_command(config_generic_models)
cli.add_command(simulation_models)
cli.add_command(download_source)
cli.add_command(transform_model)
cli.add_command(pv_deployments)
cli.add_command(prescreen_pv_penetration_levels)
cli.add_command(create_pipeline)
cli.add_command(ingest_tables)
cli.add_command(install_extensions)
cli.add_command(make_summary_tables)
cli.add_command(summarize_hosting_capacity)
cli.add_command(upgrade_cost_analysis)
cli.add_command(hosting_capacity_by_timestep)
cli.add_command(compute_hosting_capacity)
cli.add_command(plot)



from typing import Optional

from pydantic.v1 import Field

from disco.models.base import (
    BaseAnalysisModel,
    DiscoBaseModel,
    OpenDssDeploymentModel,
    SimulationModel
)


class ParameterOverridesModel(DiscoBaseModel):
    # TODO: Add upgrade parameters
    class Config:
        title = "ParameterOverridesModel"
        validate_assignment = True


class UpgradeCostAnalysisModel(BaseAnalysisModel):
    deployment: OpenDssDeploymentModel = Field(
        title="deployment",
        description="PV deployment on feeder",
        default=None,
    )
    simulation: SimulationModel = Field(
        title="simulation",
        description="Simulation parameters with PV deployment",
        default=None,
    )
    parameter_overrides: Optional[ParameterOverridesModel] = Field(
        title="parameter_overrides",
        default={},
        description="Override default upgrade parameters on job level",
    )
    
    class Config:
        title = "UpgradeCostAnalysisModel"
        anystr_strip_whitespace = True
        validate_assignment = True



"""DISCO package"""

import logging
from jade.utils.timing_utils import TimerStatsCollector

logging.getLogger(__name__).addHandler(logging.NullHandler())
timer_stats_collector = TimerStatsCollector()



import logging
import shutil
import sys
import time
from collections import namedtuple
from pathlib import Path
import json
import click
import pandas as pd

from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.extensions.generic_command import GenericCommandConfiguration
from jade.extensions.generic_command import GenericCommandParameters
from jade.loggers import setup_logging
from jade.result import ResultsSummary
from jade.utils.utils import get_cli_string, load_data, dump_data

from disco.common import EXIT_CODE_GOOD, EXIT_CODE_GENERIC_ERROR
from disco.cli.make_upgrade_tables import get_upgrade_tables, combine_job_outputs
from disco.exceptions import DiscoBaseException, get_error_code_from_exception
from disco.models.base import OpenDssDeploymentModel
from disco.models.upgrade_cost_analysis_generic_input_model import (
    UpgradeCostAnalysisSimulationModel
)
from disco.models.upgrade_cost_analysis_generic_output_model import (
    JobUpgradeSummaryOutputModel,
)
from disco.extensions.upgrade_simulation.upgrade_parameters import UpgradeParameters
from disco.extensions.upgrade_simulation.upgrade_simulation import UpgradeSimulation


logger = logging.getLogger(__name__)


JobInfo = namedtuple("JobInfo", ["name"])

AGGREGATION_JOB_NAME = "aggregate-results"


@click.command()
@click.argument("upgrades-config-file", type=click.Path(exists=True))
def check_config(upgrades_config_file):
    """Check that the upgrade cost analysis config file is valid."""
    setup_logging(__name__, None, console_level=logging.INFO, packages=["disco"])
    ret = 0
    try:
        UpgradeCostAnalysisSimulationModel.from_file(upgrades_config_file)
        print(f"UpgradeCostAnalysis config file {upgrades_config_file} is valid")
    except Exception:
        logger.exception("Failed to validate UpgradeCostAnalysis config file")
        ret = 1
    sys.exit(ret)


@click.command()
@click.argument("upgrades-config-file", type=click.Path(exists=True))
@click.option(
    "-c",
    "--config-file",
    type=str,
    default="config.json",
    show_default=True,
    help="JADE config filename",
)
@click.option(
    "--fmt",
    type=click.Choice(["csv", "json"]),
    default="json",
    show_default=True,
    help="Output data format",
)
def config(upgrades_config_file, config_file, fmt):
    """Create a JADE config file from a set of generic inputs."""
    config = UpgradeCostAnalysisSimulationModel.from_file(upgrades_config_file)
    jade_config = GenericCommandConfiguration()
    base_cmd = f"disco upgrade-cost-analysis run {upgrades_config_file} --no-aggregate-results"
    blocking_jobs = set()
    for job in config.jobs:
        cmd = f"{base_cmd} --job-name {job.name}"
        jade_job = GenericCommandParameters(
            command=cmd,
            name=job.name,
            estimated_run_minutes=job.estimated_run_minutes,
            append_output_dir=True,
        )
        jade_config.add_job(jade_job)
        blocking_jobs.add(job.name)

    aggregation_job = GenericCommandParameters(
        command=f"disco upgrade-cost-analysis aggregate-results --fmt {fmt}",
        name=AGGREGATION_JOB_NAME,
        append_output_dir=True,
        blocked_by=blocking_jobs,
    )
    jade_config.add_job(aggregation_job)

    jade_config.dump(config_file)
    print(f"Created JADE configuration file {config_file}")


def _log_level_cb(_, __, level):
    levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARNING,
        "error": logging.ERROR,
    }
    return levels[level]


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
@click.option(
    "--aggregate-results/--no-aggregate-results",
    default=True,
    show_default=True,
    help="Aggregate results from all jobs in config_file.",
)
@click.option(
    "-j",
    "--job-name",
    type=str,
    help="If set, only run the job matching this name. Otherwise, run all jobs in config_file.",
)
@click.option(
    "-o",
   "--output",
    "--jade-runtime-output",
    default="output",
    show_default=True,
    help="Output directory. If this is an absolute path, all upgraded OpenDSS master files will "
    "redirect to absolute paths of the original master files. Otherwise, relative paths will be "
    "used.",
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "--fmt",
    type=click.Choice(["csv", "json"]),
    default="json",
    show_default=True,
    help="Output data format. Only applicable if aggregate_results is true.",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite output directory if it exists.",
)
@click.option(
    "-C",
    "--console-log-level",
    type=click.Choice(["debug", "info", "warn", "error"]),
    default="info",
    show_default=True,
    help="Console log level",
    callback=_log_level_cb,
)
@click.option(
    "-F",
    "--file-log-level",
    type=click.Choice(["debug", "info", "warn", "error"]),
    default="info",
    show_default=True,
    help="File log level",
    callback=_log_level_cb,
)
def run(
    config_file,
    aggregate_results,
    job_name,
    output,
    fmt,
    force,
    console_log_level,
    file_log_level,
):
    """Run upgrade cost analysis simulation(s) from a config file."""
    jobs_output_dir = output / JOBS_OUTPUT_DIR
    jobs_output_dir.mkdir(parents=True, exist_ok=True)
    config = UpgradeCostAnalysisSimulationModel.from_file(config_file)

    if job_name is None:
        jobs = config.jobs
        log_file_dir = output
        log_filename = "run_upgrade_cost_analysis.log"
    else:
        jobs = []
        for job in config.jobs:
            if job.name == job_name:
                jobs.append(job)
                break
        if not jobs:
            print(f"Job {job_name} is not defined in {config_file}", file=sys.stderr)
            sys.exit(1)
        log_file_dir = jobs_output_dir / job_name
        log_filename = f"run_upgrade_cost_analysis__{job_name}.log"

    for job in jobs:
        _check_job_dir(jobs_output_dir / job.name, force)

    log_file_dir.mkdir(exist_ok=True)
    log_file = log_file_dir / log_filename
    setup_logging(
        __name__,
        log_file,
        console_level=console_log_level,
        file_level=file_log_level,
        packages=["disco"],
    )
    logger.info(get_cli_string())

    batch_return_code = EXIT_CODE_GOOD
    all_failed = True
    for job in jobs:
        logger.info("Run upgrades simulation for job %s", job.name)
        start = time.time()
        ret = EXIT_CODE_GOOD
        try:
            run_job(job, config, jobs_output_dir, file_log_level)
            all_failed = False
        except DiscoBaseException as exc:
            logger.exception("Unexpected DISCO error in upgrade cost analysis job=%s", job.name)
            ret = get_error_code_from_exception(type(exc))
        except Exception:
            logger.exception("Unexpected error in upgrade cost analysis job=%s", job.name)
            ret = EXIT_CODE_GENERIC_ERROR
        logger.info(
            "Completed upgrades simulation for job %s return_code=%s duration=%s seconds",
            job.name,
            ret,
            time.time() - start,
        )
        if ret != EXIT_CODE_GOOD:
            batch_return_code = ret
        _write_job_return_code(jobs_output_dir, job.name, ret)

    if aggregate_results and not all_failed:
        job_names = [x.name for x in jobs]
        _aggregate_results(output, log_file, job_names, fmt)

    # This will return an error if any job fails. If the user cares about differentiating
    # passes and failures then they should run the jobs through Jade.
    sys.exit(batch_return_code)


def _check_job_dir(job_output_dir, force):
    if job_output_dir.exists():
        if force:
            shutil.rmtree(job_output_dir)
        else:
            print(
                f"{job_output_dir} already exists. Choose a different path or set --force to overwrite.",
                file=sys.stderr,
            )
            sys.exit(1)


def _write_job_return_code(output_dir, job_name, return_code):
    _get_return_code_filename(output_dir, job_name).write_text(f"{return_code}\n")


def _read_job_return_code(output_dir, job_name):
    return int(_get_return_code_filename(output_dir, job_name).read_text().strip())


def _delete_job_return_code_file(output_dir, job_name):
    _get_return_code_filename(output_dir, job_name).unlink()


def _get_return_code_filename(output_dir, job_name):
    return output_dir / job_name / "return_code"


def run_job(job, config, jobs_output_dir, file_log_level):
    job_output_dir = jobs_output_dir / job.name
    job_output_dir.mkdir(exist_ok=True)
    job = UpgradeParameters(
        model_type="UpgradeCostAnalysisModel",
        name=job.name,
        deployment=OpenDssDeploymentModel(
            deployment_file=job.opendss_model_file,
            feeder="NA",
        ),
    )

    global_config = {
        "thermal_upgrade_params": config.thermal_upgrade_params.dict(),
        "voltage_upgrade_params": config.voltage_upgrade_params.dict(),
        "upgrade_simulation_params": {
            "enable_pydss_controller": config.enable_pydss_controllers,
        },
        "upgrade_cost_database": config.upgrade_cost_database,
        "dc_ac_ratio": config.dc_ac_ratio,
    }
    global_config["upgrade_simulation_params"]["pydss_controller"] = None
    if (config.pydss_controllers.pv_controller is not None) and config.enable_pydss_controllers:
        global_config["upgrade_simulation_params"]["pydss_controller"] = (
            config.pydss_controllers.pv_controller.dict(),
        )

    simulation = UpgradeSimulation(
        job=job,
        job_global_config=global_config,
        output=jobs_output_dir,
    )
    simulation.run(
        dc_ac_ratio=global_config["dc_ac_ratio"],
        enable_pydss_solve=global_config["upgrade_simulation_params"]["enable_pydss_controller"],
        pydss_controller_model=config.pydss_controllers.pv_controller,
        thermal_config=global_config["thermal_upgrade_params"],
        voltage_config=global_config["voltage_upgrade_params"],
        cost_database_filepath=global_config["upgrade_cost_database"],
        verbose=file_log_level == logging.DEBUG,
    )


@click.command()
@click.option(
    "-o",
    "--output",
    "--jade-runtime-output",
    required=True,
    help="Output directory",
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "--fmt",
    type=click.Choice(["csv", "json"]),
    default="json",
    show_default=True,
    help="Output data format",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def aggregate_results(output, fmt, verbose):
    """Aggregate results on a directory of upgrade cost analysis simulations."""
    level = logging.DEBUG if verbose else logging.INFO
    log_file = output / "upgrade_cost_analysis_aggregation.log"
    setup_logging(__name__, log_file, console_level=level, packages=["disco"])
    logger.info(get_cli_string())
    jade_config_file = output / CONFIG_FILE
    if not jade_config_file.exists():
        logger.error("aggregate-results is only supported when run through JADE.")
        sys.exit(1)

    job_names = (x["name"] for x in load_data(jade_config_file)["jobs"])
    _aggregate_results(output, log_file, job_names, fmt)


def _aggregate_results(output, log_file, job_names, fmt):
    jobs_output_dir = output / JOBS_OUTPUT_DIR
    output_json = {
        "results": [],
        "costs_per_equipment": [],
        "violation_summary": [],
        "equipment": [],
        "outputs": {"log_file": str(log_file), "jobs": []},
    }
    logger.info("Start result aggregation.")
    for name in job_names:
        if name == AGGREGATION_JOB_NAME:
            continue
        job_path = jobs_output_dir / name
        job_info = JobInfo(name)
        job_name = getattr(job_info, "name")
        overall_output_summary_file = job_path / "output.json"
        return_code = (
            _read_job_return_code(jobs_output_dir, name)
        )
        _delete_job_return_code_file(jobs_output_dir, name)
        if return_code != 0:
            logger.info("Skip failed job %s", name)
            continue
        data = load_data(overall_output_summary_file)
        tables = get_upgrade_tables(data)
        outputs = {
            "upgraded_opendss_model_file": str(jobs_output_dir / name / "upgraded_master.dss"),
            "return_code": return_code,
            "feeder_stats": str(jobs_output_dir / name / "feeder_stats.json"),
        }
        output_json["outputs"]["jobs"].append(outputs)
        output_json = combine_job_outputs(output_json, tables)
        
    for key in output_json: 
        if not output_json[key]:
            logger.warning("There were no aggregated %s results.", key)
    if fmt == "json":
        filename = output / "upgrade_summary.json"
        dump_data(JobUpgradeSummaryOutputModel(**output_json).dict(), filename, indent=2)
        logger.info("Output summary data to %s", filename)
    # elif fmt == "csv":
        


@click.group()
def upgrade_cost_analysis():
    """Commands related to running upgrade cost analysis simulations"""


upgrade_cost_analysis.add_command(check_config)
upgrade_cost_analysis.add_command(config)
upgrade_cost_analysis.add_command(run)
upgrade_cost_analysis.add_command(aggregate_results)



"""Implements functionality for Snapshot inputs."""

from disco.distribution.distribution_inputs import DistributionInputs


class PyDssInputs(DistributionInputs):
    """Implements functionality for PyDSS inputs."""



# -*- coding: utf-8 -*-
"""
Created on Fri Feb  5 07:46:25 2021

@author: senam
"""
import logging
import os
import json
import pandas as pd
import numpy as np

from jade.common import CONFIG_FILE
from jade.jobs.results_aggregator import ResultsAggregator
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.distribution.deployment_parameters import DeploymentParameters
from PyDSS.thermal_metrics import create_summary_from_dict
from PyDSS.pydss_project import PyDssProject
from PyDSS.pydss_results import PyDssResults
from PyDSS.node_voltage_metrics import SimulationVoltageMetricsModel, VoltageMetricsModel


logger = logging.getLogger(__name__)


def combine_metrics(project_path, scenario="control_mode"):
    summary_dict = dict()

    #if os.path.exists(project_path):
    pydss_project = PyDssProject.load_project(project_path)

    voltage_metrics = SimulationVoltageMetricsModel(
        **json.loads(
            pydss_project.fs_interface.read_file(
                os.path.join("Reports", "voltage_metrics.json")
            )
        )
    )
    assert scenario in voltage_metrics.scenarios

    data = json.loads(
            pydss_project.fs_interface.read_file(
                os.path.join("Reports", "thermal_metrics.json")
            )
        )
    thermal_metrics = create_summary_from_dict(data)
    
    voltage_summary = voltage_metrics.scenarios[scenario].summary
    summary_dict.update(voltage_summary.dict())
    summary_dict.update(thermal_metrics['summary'])
    return summary_dict

def get_absolute_changes(df, property_name, base_case='base_case'):
    to_exclude = [
        'total_num_time_points',
        'total_simulation_duration',
        'num_nodes_always_inside_ansi_a',
        'PV_capacity_kW',
        'load_capacity_kW',
        'pct_pv_to_load_ratio',
        'feeder'
    ]
    if property_name not in to_exclude:
        df[f"absolute_change_in_{property_name}"] = (
            df.loc[:, property_name] - df.loc[base_case, property_name]
        )
    return df

def aggregate_deployments(job_outputs_path, tolerance=0.05):
    main_output = os.path.dirname(job_outputs_path)
    config_file = os.path.join(main_output, CONFIG_FILE)
    config = PyDssConfiguration.deserialize(config_file)
    results = ResultsAggregator.list_results(main_output)
    result_lookup = {x.name: x for x in results}

    summary_dfs = []
    for feeder in config.list_feeders():
        all_summaries_dict = dict()
        base_case = config.get_base_case_job(feeder)
        
        for job in config.iter_feeder_jobs(feeder):
            if job.name not in result_lookup:
                logger.info("Skip missing job %s", job.name)
                continue
            if result_lookup[job.name].return_code != 0:
                logger.info("Skip failed job %s", job.name)
                continue
            logger.info("Processing %s", job.name)
            project_path = os.path.join(
                job_outputs_path, job.name, "pydss_project",
            )
            total_pv = get_total_pv_kilowatts(project_path)
            total_load = get_total_load_kilowatts(project_path)
            penetration = 100 * total_pv / max(total_load, 1e-3)
            all_summaries_dict[job.name] = {
                'feeder': feeder,
                'PV_capacity_kW': total_pv,
                'load_capacity_kW': total_load,
                'pct_pv_to_load_ratio': penetration
            }
            
            all_summaries_dict[job.name].update(combine_metrics(project_path))
            logger.info("Finished processing %s", job.name)
            
        summary_df = pd.DataFrame.from_dict(all_summaries_dict, 'index')
        if not summary_df.empty:
            for property_name in summary_df.columns:
                summary_df = get_absolute_changes(summary_df, property_name, base_case.name)
            summary_df = assess_deployments(summary_df, tolerance)
            summary_df.to_csv(os.path.join(job_outputs_path, f'impact_summary_{feeder}.csv'))
            summary_dfs.append(summary_df)

    return summary_dfs

def assess_deployments(df, tolerance):
    key = 'absolute_change_in_'
    to_exclude = {'line_max_moving_average_loading',
                  'line_num_time_points_with_instaneous_violations',
                  'line_num_time_points_with_moving_average_violations',
                  'line_instantaneous_threshold',
                  'line_moving_average_threshold',
                  'transformer_max_instantaneous_loading',
                  'transformer_max_moving_average_loading',
                  'transformer_window_size_hours',
                  'transformer_num_time_points_with_instaneous_violations',
                  'transformer_num_time_points_with_moving_average_violations',
                  'transformer_instantaneous_threshold',
                  'transformer_moving_average_threshold'
    }
    change_cols = [(c, c.split(key)[1]) for c in df.columns if key in c]
    for cols in change_cols:
        change_col = cols[0]
        col = cols[1]
        if col not in to_exclude:
            flag_col = f"pass_{col}"
            df.loc[:, flag_col] = df[change_col] <= tolerance * df[col]
    pass_flags = [c for c in df.columns if c.startswith('pass')]
    df['pass_flag'] = df[pass_flags[0]] # initialization
    for col in pass_flags[1:]:
        df['pass_flag'] = np.logical_and(df['pass_flag'], df[col])

    return df

#------------------------------------------


def get_total_pv_kilowatts(job_path):
    results = PyDssResults(job_path)
    scenario = results.scenarios[0]
    df = scenario.read_element_info_file(f'Exports/{scenario.name}/PVSystemsInfo.csv')
    if not df.Name.isnull().all():
        return df['Pmpp'].sum()
    else:
        return 0

def get_total_load_kilowatts(job_path):
    results = PyDssResults(job_path)
    scenario = results.scenarios[0]
    df = scenario.read_element_info_file(f'Exports/{scenario.name}/LoadsInfo.csv')
    
    return df['kW'].sum()
    
    
def compute_hosting_capacity(dfs, job_outputs_path):
    hc_dict = {}
    for df in dfs:
        feeder = df['feeder'].values[0]
        
        hc_dict[feeder] = {
            'max_hc_pct': None, 
            'max_hc_kW': None,
            'min_hc_pct': None, 
            'min_hc_kW': None
        }
        
        hc_set = []
        size_set = []
        if df.pass_flag.any():
            pass_df = df.loc[df.pass_flag==True, :]
            
            hc_dict[feeder]['max_hc_pct'] = pass_df.pct_pv_to_load_ratio.max()
            hc_dict[feeder]['max_hc_kW'] = pass_df.PV_capacity_kW.max()
            
            passing_pentrations = list(pass_df.pct_pv_to_load_ratio)
            for pct in passing_pentrations:
                if df.loc[df.pct_pv_to_load_ratio <= pct, 'pass_flag'].all():
                    hc_set.append(pct)
                    size_set.append(df.loc[df.pct_pv_to_load_ratio == pct, 'PV_capacity_kW'][0])
            if hc_set:
                hc_dict[feeder]['min_hc_pct'] = max(hc_set)
                hc_dict[feeder]['min_hc_kW'] = max(size_set)   
    hc_df = pd.DataFrame.from_dict(hc_dict, 'index')
    hc_df.to_csv(os.path.join(job_outputs_path, 'hosting_capacity.csv'))



import enum


class TemplateSection(enum.Enum):
    PRESCREEN = "prescreen"
    SIMULATION = "simulation"
    POSTPROCESS = "postprocess"
    MODEL = "model"
    REPORTS = "reports"


class TemplateParams(enum.Enum):
    TRANSFORM_PARAMS = "transform-params"
    CONFIG_PARAMS = "config-params"
    PRESCREEN_PARAMS = "prescreen-params"
    SUMITTER_PARAMS = "submitter-params"



from .epri.epri_model import EpriModel
from .gem.gem_model import GemModel
from .source_tree_1.source_tree_1_model import SourceTree1Model
from .source_tree_2.source_tree_2_model import SourceTree2Model
